#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Thu Apr 23 11:45:24 2020
# Author: Xiaodong Ming

"""
Independent functions
=====================

To do:
    non-object-based independent functions to support hipims IO
    
-----------------

"""
import copy
import pandas as pd
import gzip
import pickle
import os
import shutil
import glob
import scipy.signal
import numpy as np
from .rainfall_processing import _check_rainfall_rate_values

def load_object(file_name):
    """ Read a pickle file as an InputHipims/OutputHipims object
    """
    #read an object file
    try:
        with gzip.open(file_name, 'rb') as input_file:
            obj = pickle.load(input_file)
    except:
        with open(file_name, 'rb') as input_file:
            obj = pickle.load(input_file)
    print(file_name+' loaded')
    return obj

def save_object(obj, file_name, compression=True):
    """ Save an InputHipims/OutputHipims object to a pickle file 
    """
    # Overwrites any existing file.
    if compression:
        with gzip.open(file_name, 'wb') as output_file:
            pickle.dump(obj, output_file, pickle.HIGHEST_PROTOCOL)
    else:
        with open(file_name, 'wb') as output_file:
            pickle.dump(obj, output_file, pickle.HIGHEST_PROTOCOL)
    print(file_name+' has been saved')

def save_as_dict(obj, file_name):
    """Save all attributes of an input/output object to a pickle
    """
    obj_dict = copy.copy(obj.__dict__)
    if hasattr(obj, 'DEM'):
        obj_dict['DEM'] = obj.DEM.__dict__
    if hasattr(obj, 'Summary'):
        obj_dict['Summary'] = obj.Summary.to_dict()
    if hasattr(obj, 'Boundary'):
        obj_dict['Boundary'] = obj.Boundary.__dict__
    if hasattr(obj, 'Sections'):
        obj_dict.pop('Sections')
    if hasattr(obj, 'Rainfall'):
        obj_dict['Rainfall'] = obj.Rainfall.__dict__
    if hasattr(obj, 'Landcover'):
        obj_dict['Landcover'] = obj.Landcover.__dict__
    save_object(obj_dict, file_name, compression=True)

def clean_output(case_folder, num_of_sections, file_tag='*'):
    """ delete contents in output folder(s)
    """
    if case_folder[-1]!='/':
            case_folder = case_folder+'/'
    if num_of_sections==1:
        files_to_remove = case_folder+'/output/'+file_tag
        os.system('rm '+files_to_remove)
    else:    
        for i in range(num_of_sections):
            files_to_remove = case_folder+str(i)+'/output/'+file_tag
            os.system('rm '+files_to_remove)

def backup2initial(case_folder, num_of_sections=1, time_tag=None):
    """
    convert backup files from output folder to intial files of h and hU in the
    input folder
    
    name format of backup files: h_backup__3600.dat, hU_backup__3600.dat

    Parameters
    ----------
    case_folder : str, directory of the model case.
    num_of_sections : int, number of sections. The default is 1.
    time_tag : number or string, optional
        the time tag to choose backup files. The default will be the newest 
        backup file.
    Returns
    -------
    None.

    """
    if num_of_sections>1:
        for i in np.arange(num_of_sections):
            case_folder_mg = os.path.join(case_folder, str(i))
            backup2initial(case_folder_mg, 1, time_tag)
    else:
        input_folder = os.path.join(case_folder, 'input')
        output_folder = os.path.join(case_folder, 'output')
        if time_tag is None: # if tim_tag is not given, choose the newest files
            h_backupfiles = list(filter(os.path.isfile, 
                                        glob.glob(output_folder+'/h_backup__*.dat')))
            h_backupfiles.sort(key=lambda x: os.path.getmtime(x))
            hU_backupfiles = list(filter(os.path.isfile, 
                                        glob.glob(output_folder+'/hU_backup__*.dat')))
            hU_backupfiles.sort(key=lambda x: os.path.getmtime(x))
            h_backup_file = h_backupfiles[-1] # the newest h_backup file
            hU_backup_file = hU_backupfiles[-1] # the newest hU_backup file
        else:
            h_backup_file = os.path.join(output_folder, 
                                         'h_backup__'+str(time_tag)+'.dat')
            hU_backup_file = os.path.join(output_folder, 
                                         'hU_backup__'+str(time_tag)+'.dat')
        h_file = os.path.join(input_folder, 'field', 'h.dat')
        hU_file = os.path.join(input_folder, 'field', 'hU.dat')
        shutil.copy(h_backup_file, h_file)
        print(h_file+' is replaced by '+h_backup_file)
        shutil.copy(hU_backup_file, hU_file)
        print(hU_file+' is replaced by '+hU_backup_file)

#%% ***************************************************************************
# *************************Public functions************************************
def write_times_setup(case_folder=None, num_of_sections=1, time_values=None):
    """Generate a times_setup.dat file. The file contains numbers representing 
    the start time, end time, output interval, and backup interval in seconds

    Args:
        time_values: array or list of int/float, representing time in seconds, 
        default values are [0, 3600, 1800, 3600]

    """
    case_folder = _check_case_folder(case_folder)
    if time_values is None:
        time_values = np.array([0, 3600, 1800, 3600])
    time_values = np.array(time_values)
    time_values = time_values.reshape((1, time_values.size))
    if num_of_sections == 1:
        np.savetxt(case_folder+'/input/times_setup.dat', time_values, fmt='%g')
    else:
        np.savetxt(case_folder+'/times_setup.dat', time_values, fmt='%g')
    print('times_setup.dat created')

def write_device_setup(case_folder=None,
                       num_of_sections=1, device_values=None):
    """Generate a device_setup.dat file. The file contains numbers 
    representing the GPU number for each section

    Args:
        case_folder: string, the path of model
        num_of_sections: int, the number of GPUs to use
        device_values: array or list of int, representing the GPU number
    """
    case_folder = _check_case_folder(case_folder)
    if device_values is None:
        device_values = np.array(range(num_of_sections))
    device_values = np.array(device_values)
    device_values = device_values.reshape((1, device_values.size))
    if num_of_sections == 1:
        np.savetxt(case_folder+'/input/device_setup.dat',
                   device_values, fmt='%g')
    else:
        np.savetxt(case_folder+'/device_setup.dat', device_values, fmt='%g')
    print('device_setup.dat created')

def write_rain_source(rain_source, case_folder=None, num_of_sections=1):
    """ Write rainfall sources [Independent function from hipims class]

    Args:
        rain_source: numpy array, The 1st column is time in seconds, the 2nd 
        towards the end columns are rainfall rate in m/s for each source ID in 
        rainfall mask array if for multiple GPU, then copy the rain source 
        file to all domain folders
        case_folder: string, the path of model

    """
    rain_source = np.array(rain_source)
    # check rainfall source value to avoid very large raifall rates
    _ = _check_rainfall_rate_values(rain_source)
    case_folder = _check_case_folder(case_folder)
    fmt1 = '%g'  # for the first col: times in seconds
    fmt2 = '%.8e'  # for the rest array for rainfall rate m/s
    num_mask_cells = rain_source.shape[1]-1
    format_spec = [fmt2]*num_mask_cells
    format_spec.insert(0, fmt1)
    if num_of_sections == 1: # single GPU
        file_name = case_folder+'input/field/precipitation_source_all.dat'
    else: # multi-GPU
        file_name = case_folder+'0/input/field/precipitation_source_all.dat'
    with open(file_name, 'w') as file2write:
        file2write.write("%d\n" % num_mask_cells)
        np.savetxt(file2write, rain_source, fmt=format_spec, delimiter=' ')
    if num_of_sections > 1:
        for i in np.arange(1, num_of_sections):
            field_dir = case_folder+str(i)+'/input/field/'
            shutil.copy2(file_name, field_dir)
    print('precipitation_source_all.dat created')

def _write_two_arrays(file_name, id_values, bound_id_code=None):
    """Write two arrays: cell_id-value pairs and bound_id-bound_code pairs

    Inputs:

        file_name :  the full file name including path

        id_values: valid cell ID - value pair

        bound_id_code: boundary cell ID - codes pair. If bound_id_code is not 
        given, then the second part of the file won't be written (only the case
        for precipitatin_mask.dat)

    """
    if not file_name.endswith('.dat'):
        file_name = file_name+'.dat'
    if id_values.shape[1] == 3:
        fmt = ['%d %g %g']
    elif id_values.shape[1] == 2:
        fmt = ['%d %g']
    else:
        raise ValueError('Please check the shape of the 1st array: id_values')
    fmt = '\n'.join(fmt*id_values.shape[0])
    id_values_str = fmt % tuple(id_values.ravel())
    if bound_id_code is not None:
        fmt = ['%-12d %2d %2d %2d']
        fmt = '\n'.join(fmt*bound_id_code.shape[0])
        bound_id_code_str = fmt % tuple(bound_id_code.ravel())
    with open(file_name, 'w') as file2write:
        file2write.write("$Element Number\n")
        file2write.write("%d\n" % id_values.shape[0])
        file2write.write("$Element_id  Value\n")
        file2write.write(id_values_str)
        if bound_id_code is not None:
            file2write.write("\n$Boundary Numbers\n")
            file2write.write("%d\n" % bound_id_code.shape[0])
            file2write.write("$Element_id  Value\n")
            file2write.write(bound_id_code_str)

#%% static method for InputHipims
def _create_io_folders(case_folder, make_dir=False):
    """ create Input-Output path for a Hipims case
        (compatible for single/multi-GPU)
    Return:
      dir_input, dir_output, dir_mesh, dir_field
    """
    folder_name = case_folder
    dir_input = os.path.join(folder_name, 'input')
    dir_output = os.path.join(folder_name, 'output')
    if not os.path.exists(dir_output) and make_dir:
        os.makedirs(dir_output)
    if not os.path.exists(dir_input) and make_dir:
        os.makedirs(dir_input)
    dir_mesh = os.path.join(dir_input, 'mesh')
    if not os.path.exists(dir_mesh) and make_dir:
        os.makedirs(dir_mesh)
    dir_field = os.path.join(dir_input, 'field')
    if not os.path.exists(dir_field) and make_dir:
        os.makedirs(dir_field)
    data_folders = {'input':dir_input, 'output':dir_output,
                    'mesh':dir_mesh, 'field':dir_field}
    return data_folders

def _split_array_by_rows(input_array, header, split_rows, overlayed_rows=2):
    """ Clip an array into small ones according to the seperating rows

    input_array : the DEM array

    header : the DEM header

    split_rows : a list of row subscripts to split the array

    Split from bottom to top

    Return array_local, header_local: lists to store local DEM array and header

    """
    header_global = header
    end_row = header_global['nrows']-1
    overlayed_rows = 1
    array_local = []
    header_local = []
    section_sequence = np.arange(len(split_rows)+1)
    for i in section_sequence:  # from bottom to top
        section_id = i
        if section_id == section_sequence.max(): # the top section
            start_row = 0
        else:
            start_row = split_rows[i]-overlayed_rows
        if section_id == 0: # the bottom section
            end_row = header_global['nrows']-1
        else:
            end_row = split_rows[i-1]+overlayed_rows-1
        sub_array = input_array[start_row:end_row+1, :]
        array_local.append(sub_array)
        sub_header = header_global.copy()
        sub_header['nrows'] = sub_array.shape[0]
        sub_yllcorner = (header_global['yllcorner']+
                         (header_global['nrows']-1-end_row)*
                         header_global['cellsize'])
        sub_header['yllcorner'] = sub_yllcorner
        header_local.append(sub_header)
    return array_local, header_local

def _get_split_rows(input_array, num_of_sections):
    """ Split array by the number of valid cells (not NaNs) on each rows
    input_array : an array with some NaNs
    num_of_sections : (int) number of sections that the array to be splited
    return split_rows : a list of row subscripts to split the array
    Split from bottom to top
    """
    valid_cells = ~np.isnan(input_array)
    # valid_cells_count by rows
    valid_cells_count = np.sum(valid_cells, axis=1)
    valid_cells_count = np.cumsum(valid_cells_count)
    split_rows = []  # subscripts of the split row [0, 1,...]
    total_valid_cells = valid_cells_count[-1]
    for i in np.arange(num_of_sections-1): 
        num_section_cells = total_valid_cells*(i+1)/num_of_sections
        split_row = np.sum(valid_cells_count<=num_section_cells)-1
        split_rows.append(split_row)
    # sort from bottom to top
    split_rows.sort(reverse=True)
    return split_rows

def _get_cell_id_array(dem_array):
    """ to generate two arrays with the same size of dem_array:

    1. valid_id: to store valid cell id values (sequence number) starting from
    0, from bottom, left to right, top

    2. outline_id: to store valid cell id on the boundary cells

    valid_id, outline_id = __get_cell_id_array(dem_array)
    """
    # convert DEM to a two-value array: NaNs and Ones
    # and flip up and down
    dem_array_flip = np.flipud(dem_array*0+1).astype('float64')
    # Return the cumulative sum of array elements over a given axis
    # treating NaNs) as zero.
    nancumsum_vector = np.nancumsum(dem_array_flip)
    # sequence number of valid cells: 0 to number of cells-1
    valid_id = nancumsum_vector-1
    # reshape as an array with the same size of DEM
    valid_id = np.reshape(valid_id, np.shape(dem_array_flip))
    # set NaN cells as NaNs
    valid_id[np.isnan(dem_array_flip)] = np.nan
    valid_id = np.flipud(valid_id)
    # find the outline boundary cells
    array_for_outline = dem_array*0
    array_for_outline[np.isnan(dem_array)] = -1
    h_hv = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    # Convolve two array_for_outline arrays
    ind_array = scipy.signal.convolve2d(array_for_outline, h_hv, mode='same')
    ind_array[ind_array < 0] = np.nan
    ind_array[0, :] = np.nan
    ind_array[-1, :] = np.nan
    ind_array[:, 0] = np.nan
    ind_array[:, -1] = np.nan
    # extract the outline cells by a combination
    ind_array = np.isnan(ind_array) & ~np.isnan(dem_array)
    # boundary cells with valid cell id are extracted
    outline_id = dem_array*0-2 # default inner cells value:-2
    outline_id[ind_array] = 0 # outline cells:0
    return valid_id, outline_id

def _cell_subs_convertor(input_cell_subs, header_global,
                         header_local, to_global=True):
    """
    Convert global cell subs to divided local cell subs or the otherwise and 
    return output_cell_subs, only rows need to be changed

    input_cell_subs : (tuple) input rows and cols of a grid

    header_global : head information of the global grid

    header_local : head information of the local grid

    to_global : logical values, True (local to global) or False(global to local)

    Return:
        output_cell_subs: (tuple) output rows and cols of a grid
    """
    # X and Y coordinates of the centre of the first cell
    y00_centre_global = header_global['yllcorner']+\
                         (header_global['nrows']+0.5)*header_global['cellsize']
    y00_centre_local = header_local['yllcorner']+\
                        (header_local['nrows']+0.5)*header_local['cellsize']
    row_gap = (y00_centre_global-y00_centre_local)/header_local['cellsize']
    row_gap = round(row_gap)
    rows = input_cell_subs[0]
    cols = input_cell_subs[1]
    if to_global:
        rows = rows+row_gap
        # remove subs out of range of the global DEM
        ind = np.logical_and(rows >= 0, rows < header_global['nrows'])
    else:
        rows = rows-row_gap
        # remove subs out of range of the global DEM
        ind = np.logical_and(rows >= 0, rows < header_local['nrows'])
    rows = rows.astype(cols.dtype)
    rows = rows[ind]
    cols = cols[ind]
    output_cell_subs = (rows, cols)
    return output_cell_subs

#%% =======================Value check functions===============================
def _check_case_folder(case_folder):
    """ check the format of case folder
    """
    if case_folder is None:
        case_folder = os.getcwd()
    if not case_folder.endswith('/'):
        case_folder = case_folder+'/'
    return case_folder

def _check_raster_exist(file_tag):
    """check the existence of raster files .gz, .asc, .tif
    file_tag: 'DEM', 'rain_mask', 'landcover'
    """
    if os.path.isfile(file_tag+'.gz'):
        file_name = file_tag+'.gz'
    elif os.path.isfile(file_tag+'.asc'):
        file_name = file_tag+'.asc'
    elif os.path.isfile(file_tag+'.tif'):
        file_name = file_tag+'.tif'
    else:
        file_name = None
    return file_name

def _mask2dict(obj_mask, new_header=None):
    """Convert a mask Raster object to a dictionary
    """
    if new_header is not None:
        obj_mask = obj_mask.assign_to(new_header)
    array = obj_mask.array
    ind_rc = np.where(~np.isnan(array))
    mask_value = array[~np.isnan(array)]
    df = pd.DataFrame({'Row':ind_rc[0], 'Col':ind_rc[1], 'V':mask_value})
    gb = df.groupby('V')
    gb_list = [gb.get_group(x) for x in gb.groups]
    v_unique = [x.V.iloc[0] for x in gb_list]
    v_unique = np.array(v_unique)
    index = [(np.array(x.Row), np.array(x.Col)) for x in gb_list]
    mask_dict = {'value':v_unique, 'index':index}
    return mask_dict

def _dict2grid(mask_dict, array_shape):
    """Convert mask_dict to a grid array with the shape given
    """
    num_values = mask_dict['value'].size
    grid_array = np.zeros(array_shape).astype(mask_dict['value'].dtype)
    for i in np.arange(num_values):
        grid_array[mask_dict['index'][i]] = mask_dict['value'][i]
    return grid_array



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Wed Jun 24 09:20:53 2020
#Author: Xiaodong Ming

"""
Landcover
=========

To do:
    To read, compute, and show Landcover types

-------------

"""
import numpy as np
from hydro_raster.Raster import Raster
from . import indep_functions as indep_f
class Landcover:
    """ A class to set landcover data and use it to set grid parameters

    Attributes:
        mask_header: dictionary showing mask georeference
        mask_dict: dict with two keys:'value' and 'index', providing int array showing landcover type and their index respectively
        
    """
    def __init__(self, ras_data, dem_ras=None):
        """
        """
        if type(ras_data) is str:
            obj_landcover = Raster(ras_data)
        elif hasattr(ras_data, 'header'):
            obj_landcover = ras_data
        if hasattr(dem_ras, 'header'):
            # landcover resample to the same shape with DEM
            self.mask_dict = indep_f._mask2dict(obj_landcover, dem_ras.header)
            self.mask_header = dem_ras.header
            self.subs_in = np.where(~np.isnan(dem_ras.array))
        else:
            self.mask_dict = indep_f._mask2dict(obj_landcover)
            self.mask_header = obj_landcover.header
    
    def get_mask_array(self):
        """Return a mask array
        """
        array_shape = (self.mask_header['nrows'], self.mask_header['ncols'])
        mask_array = indep_f._dict2grid(self.mask_dict, array_shape)
        return mask_array
    
    def to_grid_parameter(self, param_value, land_value, default_value=0):
        """ Set grid parameter according to landcover data

        Args:
            param_value: scalar or a list of scalar
            land_ids: index representing landcover, scalar, list of scalar, 
            or list of list
        """
        mask_array = self.get_mask_array() #landcover value
        param_array = mask_array*0+default_value
        if type(param_value) is list:
            for i in np.arange(len(param_value)):
                onevalue = param_value[i]
                one_ids = land_value[i]
                ind = np.isin(mask_array, one_ids)
                param_array[ind] = onevalue
        else:
            ind = np.isin(mask_array, land_value)
            param_array[ind] = param_value
        return param_array




# -*- coding: utf-8 -*-
# @Author: Xiaodong Ming
# @Date:   2021-09-02 11:34:09
# @Last Modified by:   Xiaodong Ming
# @Last Modified time: 2022-03-03 17:55:06
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Xiaodong Ming

"""
InputHipims
===========

The module is based on the following assumptions:

    * Input DEM is a regular DEM file
    * its map unit is meter
    * its cellsize is the same in both x and y direction
    * its reference position is on the lower left corner of the southwest cell
    * All the other grid-based input files must be consistent with the DEM file

To do:

    * generate input (including sub-folder mesh and field) and output folders
    * generate mesh file (DEM.txt) and field files
    * divide model domain into small sections if multiple GPU is used

-----------------

"""

# Structure:

#    class InputHipims

#        - Initialize an object: __init__

#        - set model parameters: 

#            set_boundary_condition, set_parameter, set_rainfall, 

#            set_gauges_position, set_case_folder, set_runtime, set_device_no,

#            add_user_defined_parameter, set_num_of_sections

__author__ = "Xiaodong Ming"
import os
import shutil
import copy
import warnings
import numpy as np
from datetime import datetime
from hydro_raster.Raster import Raster
from .Boundary import Boundary
from .Rainfall import Rainfall
from .Landcover import Landcover
from .Summary import Summary
from hydro_raster.spatial_analysis import sub2map
from . import indep_functions as indep_f
#%% definition of class InputHipims
class InputHipims:
    """To define input files for a HiPIMS flood model case

    Read data, process data, write input files, and save data of a model case.

    Attributes:
        num_of_sections: (scalar) number of GPUs to run the model
        shape: shape of the DEM array
        header: (dict) header of the DEM grid
        __attributes_default: (dict) default model attribute names and values
        attributes: (dict) model attribute names and values
        times:  (list/numpy array) of four values reprenting model run time in
                seconds: start, end, output interval, backup interval
        device_no: (int) the gpu device id(s) to run model
        param_per_landcover: dict, argument to set grid parameters using the
            Landcover object. Keys are grid parameter names. Each Value is a 
            dict with three keys: param_value, land_value, default_value. 
            Refer to Landcover for more details.
        Sections: a list of objects of child-class InputHipimsSub
        Boundary: A boundary object for boundary conditions
        DEM: a Raster object to provide DEM data [alias: Raster].
        Rainfall: a Rainfall object to process rainfall data
        Landcover: a Landcover object to provide landcover data for setting
            gridded parameters        
        Summary: a Summary object to record model information
    """

#%%****************************************************************************
#***************************initialize an object*******************************
    # default parameters
    __attributes_default = {'h0':0, 'hU0x':0, 'hU0y':0,
                          'precipitation':0,
                          'precipitation_mask':0,
                          'precipitation_source':np.array([[0, 0], [1, 0]]),
                          'manning':0.035,
                          'sewer_sink':0,
                          'cumulative_depth':0, 'hydraulic_conductivity':0,
                          'capillary_head':0, 'water_content_diff':0,
                          'gauges_pos':np.array([[0, 0], [1, 1]])}
    _file_tag_list = ['z', 'h', 'hU', 'precipitation',
                      'manning', 'sewer_sink',
                      'cumulative_depth', 'hydraulic_conductivity',
                      'capillary_head', 'water_content_diff',
                      'precipitation_mask', 'precipitation_source',
                      'boundary_condition', 'gauges_pos']
    __grid_files = ['z', 'h', 'hU', 'precipitation_mask',
                  'manning', 'sewer_sink', 'precipitation',
                  'cumulative_depth', 'hydraulic_conductivity',
                  'capillary_head', 'water_content_diff']
    gridded_parameter_keys = ['h0', 'hU0x', 'hU0y', 'manning', 'sewer_sink',
                           'cumulative_depth', 'hydraulic_conductivity',
                           'capillary_head', 'water_content_diff']
    def __init__(self, dem_data=None, num_of_sections=1, case_folder=None,
                 data_path=None):
        """Initialise the InputHipims object

        Args:
            dem_data: (Raster object) or (str) provides file name of the DEM data
            data_folder: a path contain at least a DEM file named as 'DEM' with a
                suffix .gz|.asc|.tif. 'landcover' and 'rain_mask' can also be read if
                these files were given with one of the three suffix
        """
        self.attributes = InputHipims.__attributes_default.copy()
        self.num_of_sections = num_of_sections
        self.birthday = datetime.now()
        if case_folder is None:
            case_folder = os.getcwd()
        self._case_folder = case_folder
        if data_path is not None:
            self.data_path = data_path
            self._setup_by_files(data_path)
        else:
            if type(dem_data) is str:
                self.DEM = Raster(dem_data) # create Raster object
            elif type(dem_data) is Raster:
                self.DEM = dem_data
        self.Raster = self.DEM
        self.shape = self.DEM.shape
        self.header = self.DEM.header
        if not hasattr(self, 'Rainfall'):
            self.set_rainfall()
        _ = self.set_runtime()
        # get row and col index of all cells on DEM grid
        self._get_cell_subs()  # add _valid_cell_subs and _outline_cell_subs
        # divide model domain to several sections if it is not a sub section
        # each section contains a "HiPIMS_IO_class.InputHipimsSub" object
        if isinstance(self, InputHipimsSub):
            pass
        else:
            self.__divide_grid()
        self.set_case_folder() # set data_folders
        self.set_device_no() # set the device number
        self.set_boundary_condition(outline_boundary='fall')
        self.set_gauges_position()
        self._initialize_summary_obj()# initialize a Model Summary object
        
    def __str__(self):
        """To show object summary information when it is called in console
        """
        self.Summary.display()
        time_str = self.birthday.strftime('%Y-%m-%d %H:%M:%S')
        return  self.__class__.__name__+' object created on '+ time_str
#    __repr__ = __str__
#******************************************************************************
#%%**********************Get object attributes**************************************
    def get_case_folder(self):
        """Get the directory of the case.

        Return: 
            string: The directory of the case
        """
        return self._case_folder

    def get_data_folders(self):
        """Get the directories and subdirectories of the input and output data.
        
        Return:
            string: a dict with four keys ('input', 'output', 'mesh', 'field')
                providing their absolute paths
        """
        return self._data_folders
    
    def get_input_filenames(self):
        """Get the input file names.
        
        Return:
            list: a list of string giving all input filenames without 
        """
        return self._file_tag_list

#%%**********************Setup the object**************************************
    def set_boundary_condition(self, boundary_list=None,
                               outline_boundary='fall'):
        """To create a Boundary object for boundary conditions
        
        Args:
            boundary_list: (list of dicts), each dict contain keys (polyPoints,
                type, h, hU) to define a IO boundary's position, type, and
                Input-Output (IO) sources timeseries. Keys including:

                1.polyPoints is a numpy array giving X(1st col) and Y(2nd col)
                    coordinates of points to define the position of a boundary.
                    An empty polyPoints means outline boundary.

                2.type: 'open'(flow out flatly), 'rigid'(no outlet),
                        'fall'(water flow out like a fall)

                3.h: a two-col numpy array. The 1st col is time(s). The 2nd col
                     is water depth(m)

                4.hU: a two-col numpy array. The 1st col is time(s). The 2nd
                    col is discharge(m3/s) or a three-col numpy array, the 2nd
                    col and the 3rd col are velocities(m/s) in x and y
                    direction, respectively.
            outline_boundary: (str) 'open'|'rigid', default outline boundary is
                open and both h and hU are set as zero

        """
        if boundary_list is None and hasattr(self, 'Boundary'):
            boundary_list = self.Boundary.boundary_list
        bound_obj = Boundary(boundary_list, outline_boundary)
        valid_subs = self._valid_cell_subs
        outline_subs = self._outline_cell_subs
        if not isinstance(self, InputHipimsSub):
            dem_header = self.header
        # add the subsripts and id of boundary cells on the domain grid
            bound_obj._fetch_boundary_cells(valid_subs,
                                            outline_subs, dem_header)
            bound_obj._convert_flow2velocity(self.DEM)
        self.Boundary = bound_obj
        if hasattr(self, 'Sections'):
            bound_obj._divide_domain(self)       
        if hasattr(self, 'Summary'):
            self.Summary.set_boundary_attr(self.Boundary)

    def set_initial_condition(self, parameter_name, parameter_value):
        """Set initial condition for h0, hU0x, hU0y
        
        Args:
            parameter_name: (str) h0, hU0x, hU0y
            parameter_value: scalar or numpy array with the same size of DEM.
        """
        if parameter_name not in ['h0', 'hU0x', 'hU0y']:
            raise ValueError('Parameter is not recognized: '+parameter_name)
        if type(parameter_value) is np.ndarray:
            if parameter_value.shape != self.shape:
                raise ValueError('The array of the parameter '
                                 'value should have the same '
                                 'shape with the DEM array')
        elif np.isscalar(parameter_value) is False:
            raise ValueError('The parameter value must be either '
                             'a scalar or an numpy array')
        self.attributes[parameter_name] = parameter_value
        self.Summary.set_initial_attr(**{parameter_name:parameter_value})
    
    def set_grid_parameter(self, **kwargs):
        """ Set grid parameter with Landcover object as name=value

        Args:
            kwargs: Keyword Arguments Specified by a Dictionary
                keyword: name, from grid_parameter_keys

                value:  1. scalar, a uniform parameter value
                        2. array, gridded parameter value with the same shape
                            of the DEM array
                        3. dict, contains param_value, land_value,  and 
                            default_value=0
        """
        if not hasattr(self, 'param_per_landcover'):
            # save arguments to call Landcover.to_grid_parameter()
            self.param_per_landcover = {}
        for keyword, value in kwargs.items():
            if keyword not in self.gridded_parameter_keys:
                raise ValueError('Parameter is not recognized: '+keyword)
            if type(value) is np.ndarray:
                if value.shape != self.shape:
                    raise ValueError('The array of the parameter '+keyword+
                           ' should have the same shape with the DEM array')
                else:
                    self.attributes[keyword] = value
            elif np.isscalar(value):
                self.attributes[keyword] = value
            elif type(value) is dict:
                self.param_per_landcover[keyword] = value
                value_array = self.Landcover.to_grid_parameter(**value)
                self.attributes[keyword] = value_array
            else:
                raise ValueError(keyword+' must be a scalar, array or dict')
            self.Summary.set_params_attr(**{keyword:value})

    def set_rainfall(self, rain_mask=None, rain_source=None):
        """ Set rainfall mask and rainfall source

        Args:
            rain_mask: str [filename of a Raster endswith .gz/asc/tif]
                    numpy int array with the same shape with DEM array
                    a Raster object
            rain_source: str [filename of a csv file for rainfall source data]
                    numpy array the 1st column is time in seconds, 2nd to
                    the end columns are rainfall rates in m/s.
        """
        if not hasattr(self, 'Rainfall'):
            self.Rainfall = Rainfall(self.attributes['precipitation_mask'], 
                                     self.attributes['precipitation_source'],
                                     dem_ras=self.DEM)
        # set rain mask
        if (rain_mask is None) & (rain_source is None): 
            # use default or pre-defined value
            rain_mask = self.attributes['precipitation_mask']
            rain_source = self.attributes['precipitation_source']
            self.Rainfall = Rainfall(rain_mask, rain_source, dem_ras=self.DEM)
        if rain_mask is not None:
            self.Rainfall.set_mask(rain_mask, dem_ras=self.DEM)
        if rain_source is not None:
            self.Rainfall.set_source(rain_source)
        if hasattr(self, 'Summary'):
            self.Summary.rain_attr = self.Rainfall.attrs

    def set_gauges_position(self, gauges_pos=None):
        """Set coordinates of monitoring gauges

        Args:
            gauges_pos: (numpy array) the 1st column is X coordinates, 
                2nd column is the Y coordinates
        """
        if gauges_pos is None:
            gauges_pos = self.attributes['gauges_pos']
        if type(gauges_pos) is list:
            gauges_pos = np.array(gauges_pos)
        if gauges_pos.shape[1] != 2:
            raise ValueError('The gauges_pos arraymust have two columns')
        self.attributes['gauges_pos'] = gauges_pos
#        self.Summary.add_param_infor('gauges_pos', gauges_pos)
        # for multi_GPU, divide gauges based on the extent of each section
        if hasattr(self, 'Sections'):
            pos_X = gauges_pos[:,0]
            pos_Y = gauges_pos[:,1]
            for obj_section in self.Sections:
                extent = obj_section.DEM.extent
                ind_x = np.logical_and(pos_X >= extent[0], pos_X <= extent[1])
                ind_y = np.logical_and(pos_Y >= extent[2], pos_Y <= extent[3])
                ind = np.where(np.logical_and(ind_x, ind_y))
                ind = ind[0]
                obj_section.attributes['gauges_pos'] = gauges_pos[ind,:]
                obj_section.attributes['gauges_ind'] = ind
        if hasattr(self, 'Summary'):
            self.Summary.set_model_attr(gauges_pos=gauges_pos)

    def set_case_folder(self, new_folder=None, make_dir=False):
        """ Initialize, renew, or create case and data folders
        
        Args:
            new_folder: (str) renew case and data folder if it is given
            make_dir: True|False create folders if it is True
        """
        # to change case_folder
        if new_folder is None:
            new_folder = self._case_folder
        self._case_folder = new_folder
        self._data_folders = indep_f._create_io_folders(self._case_folder,
                                               make_dir)
        # for multiple GPUs
        if hasattr(self, 'Sections'):
            for obj in self.Sections:
                sub_case_folder = os.path.join(new_folder, str(obj.section_id))
                obj.set_case_folder(sub_case_folder)                        
        if hasattr(self, 'Summary'):
            self.Summary.set_model_attr(case_folder=self._case_folder)

    def set_runtime(self, runtime=None):
        """set runtime of the model

        Args:
            runtime: a list of four values representing start, end, output 
                interval and backup interval respectively. If only three
                were given, then the backup interval will be given as the end
                time(the 2nd value)
        """
        if runtime is None:
            runtime = [0, 3600, 3600, 3600]
        runtime = np.array(runtime)
        if runtime.size == 3: #
            runtime = np.insert(runtime, 3, runtime[1])
        self.times = runtime
        runtime_str = ('{0}-start, {1}-end, {2}-output interval, '
                       '{3}-backup interval')
        runtime_str = runtime_str.format(*runtime)
        if hasattr(self, 'Summary'):
#            self.Summary.add_items('Runtime(s)', runtime_str)
            self.Summary.set_model_attr(run_time=self.times)
        return runtime_str

    def set_device_no(self, device_no=None):
        """set device no of the model

        Args:
            device_no: int or a list of int corresponding to the number of 
                sections 
        """
        if device_no is None:
            device_no = np.arange(self.num_of_sections)
        device_no = np.array(device_no)
        self.device_no = device_no
    
    def set_landcover(self, landcover_data):
        """ Set Landcover object with a Raster object or file

        Args:
            landcover_data: (string or Raster) A Raster file or its file name 
                of land cover data
        """
        self.Landcover = Landcover(landcover_data, self.DEM)

    def add_user_defined_parameter(self, param_name, param_value):
        """ Add a grid-based user-defined parameter to the model

        Args:
            param_name: (str) name the parameter and the input file name too
            param_value: (scalar) or (numpy arary) with the same shape of the 
                DEM array
        """
        if param_name not in InputHipims.__grid_files:
            InputHipims.__grid_files.append(param_name)
            InputHipims._file_tag_list.append(param_name)
        self.attributes[param_name] = param_value
        print(param_name+ ' is added to the InputHipims object')
    
    def set_num_of_sections(self, num_of_sections):
        """ set the number of divided sections to run a case

        It can transfer single-gpu to multi-gpu or the opposite way

        Args:
            num_of_sections: (int) number of domains
        """
        self.num_of_sections = num_of_sections
        self.set_device_no()
        if num_of_sections==1: # to single GPU
            if hasattr(self, 'Sections'):
                del self.Sections
        else: # to multiple GPU
            self.__divide_grid()
            outline_boundary = self.Boundary.outline_boundary
            self.set_boundary_condition(outline_boundary=outline_boundary)
            self.set_gauges_position()
            self.Boundary._divide_domain(self)
        self.set_case_folder(self._case_folder)
        self.birthday = datetime.now()
        self.Summary.set_model_attr(num_GPU=self.num_of_sections)
#        time_str = self.birthday.strftime('%Y-%m-%d %H:%M:%S')


#%%****************************************************************************
#************************Write input files*************************************
    def write_input_files(self, file_tag=None, ignore_one_value_files=False):
        """ Write input files

        To classify the input files and call functions needed to write each
            input files
        
        Args:
            file_tag: a string or list of string giving the name(s) of input
                files without suffix
            ignore_one_value_files: do not write one value grid files if it
                is true, including manning, sewer_sink, cumulative_depth,
                hydraulic_conductivity, capillary_head, water_content_diff
        """
        ignore_file_lists = ['manning', 'sewer_sink',
                  'cumulative_depth', 'hydraulic_conductivity',
                  'capillary_head', 'water_content_diff']
        self._make_data_dirs()
        
        if ignore_one_value_files:
            for one_file in ignore_file_lists:
                file_value = np.array(self.attributes[one_file])
                if file_value.size > 1: # remove files with more than one value
                    ignore_file_lists.remove(one_file)
            # rectify grid file list by removing the one value files
            grid_files = [x for x in InputHipims.__grid_files if x not in ignore_file_lists]
        else:
            grid_files = InputHipims.__grid_files

        if file_tag is None or file_tag == 'all': # write all files
            write_list = self._file_tag_list
            if self.num_of_sections > 1:
                self.write_halo_file()
            self.write_mesh_file()
            self.write_runtime_file()
            self.write_device_file()
            self.write_default_param_file()
        elif type(file_tag) is str:
            write_list = [file_tag]
        elif type(file_tag) is list:
            write_list = file_tag
        else:
            print(self._file_tag_list)
            raise ValueError(('file_tag should be a string or a list of string'
                             'from the above list'))
        for one_file in write_list:
            if one_file in grid_files: # grid-based files
                self.write_grid_files(one_file)
            elif one_file == 'boundary_condition':
                self.write_boundary_conditions()
            elif one_file == 'precipitation_source':
                self.write_rainfall_source()
            elif one_file == 'gauges_pos':
                self.write_gauges_position()
            else:
                print(one_file+' is ignored')
                #raise ValueError(one_file+' is not recognized')

    def write_default_param_file(self):
        """ write default_param.dat to provide default values of grid-based
        parameters
        """
        default_params = {
                  'precipitation':0,
                  'manning':0.035,
                  'sewer_sink':0,
                  'cumulative_depth':0, 
                  'hydraulic_conductivity':0,
                  'capillary_head':0,
                  'water_content_diff':0}
        for key, value in default_params.items():
            file_value = np.array(self.attributes[key])
            if file_value.size == 1: # remove files with more than one value
                default_params[key] = self.attributes[key]
        case_folder = self.get_case_folder()
        if hasattr(self, 'Sections'):
             for obj_section in self.Sections:
                 obj_section.write_default_param_file()
        else:
            file_name = os.path.join(case_folder, 'input', 'field', 'default_param.dat')
            with open(file_name, 'w') as file:
                for key, value in default_params.items():
                    file.write(key+' '+str(value)+'\n')
        

    def write_grid_files(self, file_tag, is_single_gpu=False):
        """Write grid-based files

        Public version for both single and multiple GPUs

        Args:
            file_tag: the pure name of a grid-based file
        """
        self._make_data_dirs()
        grid_files = InputHipims.__grid_files
        if file_tag not in grid_files:
            raise ValueError(file_tag+' is not a grid-based file')
        if is_single_gpu or self.num_of_sections == 1:
            # write as single GPU even the num of sections is more than one
            self._write_grid_files(file_tag, is_multi_gpu=False)
        else:
            self._write_grid_files(file_tag, is_multi_gpu=True)
        readme_filename = os.path.join(self._case_folder,'readme.txt')
        self.Summary.to_json(readme_filename)
        print(file_tag+' created')

    def write_boundary_conditions(self):
        """ Write boundary condtion files

        if there are multiple domains, write in the first folder
            and copy to others
        """
        self._make_data_dirs()
        if self.num_of_sections > 1:  # multiple-GPU
            field_dir = self.Sections[0]._data_folders['field']
            file_names_list = self.__write_boundary_conditions(field_dir)
            self.__copy_to_all_sections(file_names_list)
        else:  # single-GPU
            field_dir = self._data_folders['field']
            self.__write_boundary_conditions(field_dir)
        readme_filename = os.path.join(self._case_folder,'readme.txt')
        self.Summary.to_json(readme_filename)
        print('boundary condition files created')

    def write_rainfall_source(self):
        """Write rainfall source data

        rainfall mask can be written by function write_grid_files
        """
        self._make_data_dirs()
        if hasattr(self, 'Rainfall'):
            rain_source = self.Rainfall.get_source_array()
        else:
            rain_source = self.attributes['precipitation_source']
        case_folder = self._case_folder
        num_of_sections = self.num_of_sections
        indep_f.write_rain_source(rain_source, case_folder, num_of_sections)
        readme_filename = os.path.join(self._case_folder,'readme.txt')
        self.Summary.to_json(readme_filename)

    def write_gauges_position(self, gauges_pos=None):
        """ Write the gauges position file

        Public version for both single and multiple GPUs

        Args:
            gauge_pos: (2-coloumn numpy array) positions of gauges 
        """
        self._make_data_dirs()
        if gauges_pos is not None:
            self.set_gauges_position(np.array(gauges_pos))
        if self.num_of_sections > 1:  # multiple-GPU
            for obj_section in self.Sections:
                field_dir = obj_section._data_folders['field']
                obj_section.__write_gauge_ind(field_dir)
                obj_section.__write_gauge_pos(field_dir)
        else:  # single-GPU
            field_dir = self._data_folders['field']
            self.__write_gauge_pos(field_dir)
        readme_filename = os.path.join(self._case_folder,'readme.txt')
        self.Summary.to_json(readme_filename)
        print('gauges_pos.dat created')

    def write_halo_file(self):
        """ Write overlayed cell IDs
        """
        num_section = self.num_of_sections
        case_folder = self._case_folder
        file_name = os.path.join(case_folder, 'halo.dat')
        with open(file_name, 'w') as file2write:
            file2write.write("No. of Domains\n")
            file2write.write("%d\n" % num_section)
            for obj_section in self.Sections:
                file2write.write("#%d\n" % obj_section.section_id)
                overlayed_id = obj_section.overlayed_id
                for key in ['bottom_low', 'bottom_high',
                            'top_high', 'top_low']:
                    if key in overlayed_id.keys():
                        line_ids = overlayed_id[key]
                        line_ids = np.reshape(line_ids, (1, line_ids.size))
                        np.savetxt(file2write,
                                   line_ids, fmt='%d', delimiter=' ')
                    else:
                        file2write.write(' \n')
        print('halo.dat created')

    def write_mesh_file(self, is_single_gpu=False):
        """ Write mesh file DEM.txt, compatoble for both single and multiple
            GPU model
        """
        self._make_data_dirs()
        if is_single_gpu is True or self.num_of_sections == 1:
            file_name = os.path.join(self._data_folders['mesh'],
                                     'DEM.txt')
            self.DEM.write_asc(file_name)
        else:
            for obj_section in self.Sections:
                file_name = os.path.join(obj_section._data_folders['mesh'],
                                         'DEM.txt')
                obj_section.DEM.write_asc(file_name)
        readme_filename = os.path.join(self._case_folder,'readme.txt')
        self.Summary.to_json(readme_filename)
    
    def write_runtime_file(self, time_values=None):
        """ write times_setup.dat file
        """
        if time_values is None:
            time_values = self.times
        indep_f.write_times_setup(self._case_folder, self.num_of_sections,
                                  time_values)
        self.Summary.to_json(self._case_folder+'/readme.txt')
    
    def write_device_file(self, device_no=None):
        """Create device_setup.dat for choosing GPU number to run the model
        """
        if device_no is None:
            device_no = self.device_no
        indep_f.write_device_setup(self._case_folder, self.num_of_sections,
                                   device_no)

    def save_object(self, file_name):
        """ Save object as a pickle file
        """
        indep_f.save_as_dict(self, file_name)

#%%****************************************************************************
#******************************* Visualization ********************************
    def domain_show(self, title='Domain Map', show_split=False, **kwargs):
        """Show domain map of the object
        
        Args:
            title: string to set the figure titile, 'Domain Map' is the defualt
            show_split: logical, show split lines for multi-gpu division 
        
        Retrun:
            fig: figure handle
            ax: axis handle
        """
        obj_dem = copy.deepcopy(self.DEM)
        fig, ax = obj_dem.mapshow(title=title, cax_str='DEM(m)', **kwargs)
        cell_subs = self.Boundary.cell_subs
        legends = []
        if show_split:            
            if hasattr(self, 'Sections'):
                x_overlayed = []
                y_overlayed = []
                for obj_sub in self.Sections:
                    overlayed_subs = obj_sub.overlayed_cell_subs_global
                    rows = overlayed_subs[0]
                    cols = overlayed_subs[1]
                    X, Y = sub2map(rows, cols, self.DEM.header)
                    x_overlayed = np.append(x_overlayed, X)
                    y_overlayed = np.append(y_overlayed, Y)
                ax.plot(x_overlayed, y_overlayed, '.k')
                legends.append('Splitting cells')
            else:
                warnings.warn("show_split is only for multi-gpu model")
        num = 0
        for cell_sub in cell_subs:
            rows = cell_sub[0]
            cols = cell_sub[1]
            X, Y = sub2map(rows, cols, self.DEM.header)
            ax.plot(X, Y, '.')
            if num==0:
                legends.append('Outline cells')
            else:
                legends.append('Boundary '+str(num))
            num = num+1
        
        ax.legend(legends, edgecolor=None, facecolor=None, loc='best',
                  fontsize='x-small')
        return fig, ax
    
    def plot_rainfall_map(self, figname=None, method='sum', **kw):
        """plot rainfall map within model domain
        """
        fig, ax = self.Rainfall.plot_rainfall_map(method='sum', **kw)
        cell_subs = self._outline_cell_subs
        rows = cell_subs[0]
        cols = cell_subs[1]
        X, Y = sub2map(rows, cols, self.DEM.header)
        ax.plot(X, Y, '.k')
        return fig, ax 

    def plot_rainfall_curve(self, start_date=None, method='mean', **kw):
        """ Plot time series of average rainfall rate inside the model domain

        Args:
            start_date: a datetime object to give the initial datetime of rain
            method: 'mean'|'max','min','mean', method to calculate gridded
                rainfall over the model domain
        """
        fig, ax = self.Rainfall.plot_time_series(method, **kw)
        return fig, ax

#%%****************************************************************************
#*************************** Protected methods ********************************
    def _get_cell_subs(self, dem_array=None):
        """ To get valid_cell_subs and outline_cell_subs for the object

        To get the subscripts of each valid cell on grid
        """
        if dem_array is None:
            dem_array = self.DEM.array
        valid_id, outline_id = indep_f._get_cell_id_array(dem_array)
        subs = np.where(~np.isnan(valid_id))
        id_vector = valid_id[subs]
        # sort the subscripts according to cell id values
        sorted_vectors = np.c_[id_vector, subs[0], subs[1]]
        sorted_vectors = sorted_vectors[sorted_vectors[:, 0].argsort()]
        self._valid_cell_subs = (sorted_vectors[:, 1].astype('int32'),
                                 sorted_vectors[:, 2].astype('int32'))
        subs = np.where(outline_id == 0) # outline boundary cell
        outline_id_vect = outline_id[subs]
        sorted_array = np.c_[outline_id_vect, subs[0], subs[1]]
        self._outline_cell_subs = (sorted_array[:, 1].astype('int32'),
                                   sorted_array[:, 2].astype('int32'))
    def __divide_grid(self):
        """Divide DEM grid to sub grids

        Create objects based on sub-class InputHipimsSub
        """
        if isinstance(self, InputHipimsSub):
            return 0  # do not divide InputHipimsSub objects, return a number
        else:
            if self.num_of_sections == 1:
                return 1 # do not divide if num_of_sections is 1
        num_of_sections = self.num_of_sections
        dem_header = self.header
        # subscripts of the split row [0, 1,...] from bottom to top
        split_rows = indep_f._get_split_rows(self.DEM.array, num_of_sections)
        array_local, header_local = \
            indep_f._split_array_by_rows(self.DEM.array, dem_header,
                                         split_rows)
        # to receive InputHipimsSub objects for sections
        Sections = []
        section_sequence = np.arange(num_of_sections)
        header_global = dem_header
        for i in section_sequence:  # from bottom to top
            case_folder = os.path.join(self._case_folder, str(i))
            # create a sub object of InputHipims
            sub_hipims = InputHipimsSub(array_local[i], header_local[i],
                                        case_folder, num_of_sections)
            # get valid_cell_subs on the global grid
            valid_cell_subs = sub_hipims._valid_cell_subs
            valid_subs_global = \
                 indep_f._cell_subs_convertor(valid_cell_subs, header_global,
                                      header_local[i], to_global=True)
            sub_hipims.valid_subs_global = valid_subs_global
            # record section sequence number
#            sub_hipims.section_id = i
            #get overlayed_id (top two rows and bottom two rows)
            top_h = np.where(valid_cell_subs[0] == 0)
            top_l = np.where(valid_cell_subs[0] == 1)
            bottom_h = np.where(
                valid_cell_subs[0] == valid_cell_subs[0].max()-1)
            bottom_l = np.where(valid_cell_subs[0] == valid_cell_subs[0].max())
            if i == 0: # the bottom section
                overlayed_id = {'top_high':top_h[0], 'top_low':top_l[0]}
            elif i == self.num_of_sections-1: # the top section
                overlayed_id = {'bottom_low':bottom_l[0],
                                'bottom_high':bottom_h[0]}
            else:
                overlayed_id = {'top_high':top_h[0], 'top_low':top_l[0],
                                'bottom_high':bottom_h[0],
                                'bottom_low':bottom_l[0]}
            sub_hipims.overlayed_id = overlayed_id
            all_ids = list(overlayed_id.values())
            all_ids = np.concatenate(all_ids).ravel()
            all_ids.sort()
            overlayed_cell_subs_global = (valid_subs_global[0][all_ids],
                                          valid_subs_global[1][all_ids])
            sub_hipims.overlayed_cell_subs_global = overlayed_cell_subs_global
            Sections.append(sub_hipims)
        # reset global var section_id of InputHipimsSub
        InputHipimsSub.section_id = 0
        self.Sections = Sections
        self._initialize_summary_obj()# get a Model Summary object

    def _get_vector_value(self, attribute_name, is_multi_gpu=True,
                          add_initial_water=True):
        """ Generate a single vector for values in each grid cell sorted based
                on cell IDs
        
        Args:
            attribute_name: attribute names based on a grid

        Return:
            output_vector: a vector or list vectors, giving values on valid 
                global cells or sub-domain cells
        """
        # get grid value
        dem_shape = self.shape
        grid_values = np.zeros(dem_shape)
        # set grid value for the entire domain
        if attribute_name == 'z':
            grid_values = self.DEM.array
        elif attribute_name == 'h':
            grid_values = grid_values+self.attributes['h0']
            # traversal each boundary to add initial water
            if add_initial_water:
                wet_cell_subs = self.Boundary.cell_subs_wet_io
                if wet_cell_subs is not None:
                    grid_values[wet_cell_subs] = grid_values[wet_cell_subs]+0.001
        elif attribute_name == 'hU':
            grid_values0 = grid_values+self.attributes['hU0x']
            grid_values1 = grid_values+self.attributes['hU0y']
            grid_values = [grid_values0, grid_values1]
        elif attribute_name == 'precipitation_mask':
            if hasattr(self, 'Rainfall'):
                grid_values = self.Rainfall.get_mask_array()
            else:
                grid_values = grid_values+self.attributes[attribute_name]
        else:
            if hasattr(self, 'param_per_landcover'):
                arg_dicts = self.param_per_landcover
                if attribute_name in arg_dicts.keys():
                    arg_dict = arg_dicts[attribute_name]
                    grid_values = self.Landcover.to_grid_parameter(**arg_dict)
                else:
                    grid_values = grid_values+self.attributes[attribute_name]
            else:
                grid_values = grid_values+self.attributes[attribute_name]

        def grid_to_vect(grid_values, cell_subs):
            """ Convert grid values to 1 or 2 col vector values
            """
            if type(grid_values) is list:
                vector_value0 = grid_values[0][cell_subs]
                vector_value1 = grid_values[1][cell_subs]
                vector_value = np.c_[vector_value0, vector_value1]
            else:
                vector_value = grid_values[cell_subs]
            return vector_value
        #
        if is_multi_gpu: # generate vector value for multiple GPU
            output_vector = []
            for obj_section in self.Sections:
                cell_subs = obj_section.valid_subs_global
                vector_value = grid_to_vect(grid_values, cell_subs)
                output_vector.append(vector_value)
        else:
            output_vector = grid_to_vect(grid_values, self._valid_cell_subs)
        return output_vector

    def _get_boundary_id_code_array(self, file_tag='z'):
        """To generate a 4-col array of boundary cell id (0) and code (1~3)
        """
        bound_obj = self.Boundary
        output_array_list = []
        for ind_num in np.arange(bound_obj.num_of_bound):
            if file_tag == 'h':
                bound_code = bound_obj.data_table.h_code[ind_num]
            elif file_tag == 'hU':
                bound_code = bound_obj.data_table.hU_code[ind_num]
            else:
                bound_code = np.array([[2, 0, 0]]) # shape (1, 3)
            if bound_code.ndim < 2:
                bound_code = np.reshape(bound_code, (1, bound_code.size))
            cell_id = bound_obj.cell_id[ind_num]
            if cell_id.size > 0:
                bound_code_array = np.repeat(bound_code, cell_id.size, axis=0)
                id_code_array = np.c_[cell_id, bound_code_array]
                output_array_list.append(id_code_array)
        # add overlayed cells with [4, 0, 0]
        # if it is a sub section object, there should be attributes:
        # overlayed_id, and section_id
        if hasattr(self, 'overlayed_id'):
            cell_id = []
            if 'top_high' in self.overlayed_id:
                cell_id += self.overlayed_id['top_high'].tolist()
            if 'bottom_low' in self.overlayed_id:
                cell_id += self.overlayed_id['bottom_low'].tolist()
            cell_id = np.array(cell_id)
            bound_code = np.array([[4, 0, 0]]) # shape (1, 3)
            bound_code_array = np.repeat(bound_code, cell_id.size, axis=0)
            id_code_array = np.c_[cell_id, bound_code_array]
            output_array_list.append(id_code_array)
        output_array = np.concatenate(output_array_list, axis=0)
        # when unique the output array according to cell id
        # keep the last occurrence rather than the default first occurrence
        output_array = np.flipud(output_array) # make the IO boundaries first
        _, ind = np.unique(output_array[:, 0], return_index=True)
        output_array = output_array[ind]
        return output_array

    def _initialize_summary_obj(self):
        """ Initialize the model summary object
        """
        summary_obj = Summary(self)
        self.Summary = summary_obj

    def _write_grid_files(self, file_tag, is_multi_gpu=True):
        """ Write input files consistent with the DEM grid

        Private function called by public function write_grid_files
        file_name: includes ['h','hU','precipitation_mask',
                             'manning','sewer_sink',
                             'cumulative_depth', 'hydraulic_conductivity',
                             'capillary_head', 'water_content_diff']
        """
        if is_multi_gpu is True:  # write for multi-GPU, use child object
            vector_value_list = self._get_vector_value(file_tag, is_multi_gpu)
            for obj_section in self.Sections:
                vector_value = vector_value_list[obj_section.section_id]
                cell_id = np.arange(vector_value.shape[0])
                cells_vect = np.c_[cell_id, vector_value]
                file_name = os.path.join(obj_section._data_folders['field'],
                                         file_tag+'.dat')
                if file_tag == 'precipitation_mask':
                    bounds_vect = None
                else:
                    bounds_vect = \
                        obj_section._get_boundary_id_code_array(file_tag)
                indep_f._write_two_arrays(file_name, cells_vect, bounds_vect)
        else:  # single GPU, use global object
            file_name = os.path.join(self._data_folders['field'],
                                     file_tag+'.dat')
            vector_value = self._get_vector_value(file_tag, is_multi_gpu=False)
            cell_id = np.arange(vector_value.shape[0])
            cells_vect = np.c_[cell_id, vector_value]
            if file_tag == 'precipitation_mask':
                bounds_vect = None
            else:
                bounds_vect = self._get_boundary_id_code_array(file_tag)
            indep_f._write_two_arrays(file_name, cells_vect, bounds_vect)
        return None

    def _make_data_dirs(self):
        """ Create folders in current device
        """
        if hasattr(self, 'Sections'):
            for obj_section in self.Sections:
                indep_f._create_io_folders(obj_section.get_case_folder(),
                                           make_dir=True)
        else:
            indep_f._create_io_folders(self._case_folder, make_dir=True)
    
    def _dict2grid(self, mask_dict):
        """Convert mask_dict to a grid array with the same shape of DEM
        """
        num_values = mask_dict['value'].size
        grid_array = np.zeros(self.DEM.shape).astype(mask_dict['value'].dtype)
        for i in np.arange(num_values):
            grid_array[mask_dict['index'][i]] = mask_dict['value'][i]
        return grid_array
        
#------------------------------------------------------------------------------
#*************** Private methods only for the parent class ********************
#------------------------------------------------------------------------------
    def __write_boundary_conditions(self, field_dir, file_tag='both'):
        """ Write boundary condition source files,if hU is given as flow
        timeseries, convert flow to hUx and hUy.
        Private function to call by public function write_boundary_conditions
        file_tag: 'h', 'hU', 'both'
        h_BC_[N].dat, hU_BC_[N].dat
        if hU is given as flow timeseries, convert flow to hUx and hUy
        """
        obj_boundary = self.Boundary
        file_names_list = []
        fmt_h = ['%g', '%g']
        fmt_hu = ['%g', '%g', '%g']
        # write h_BC_[N].dat
        if file_tag in ['both', 'h']:
            h_sources = obj_boundary.data_table['hSources']
            ind_num = 0
            for i in np.arange(obj_boundary.num_of_bound):
                h_source = h_sources[i]
                if h_source is not None:
                    file_name = os.path.join(field_dir,
                                             'h_BC_'+str(ind_num)+'.dat')
                    np.savetxt(file_name, h_source, fmt=fmt_h, delimiter=' ')
                    ind_num = ind_num+1
                    file_names_list.append(file_name)
        # write hU_BC_[N].dat
        if file_tag in ['both', 'hU']:
            hU_sources = obj_boundary.data_table['hUSources']
            ind_num = 0
            for i in np.arange(obj_boundary.num_of_bound):
                hU_source = hU_sources[i]
                if hU_source is not None:
                    file_name = os.path.join(field_dir,
                                             'hU_BC_'+str(ind_num)+'.dat')
                    np.savetxt(file_name, hU_source, fmt=fmt_hu, delimiter=' ')
                    ind_num = ind_num+1
                    file_names_list.append(file_name)
        return file_names_list

    def __write_gauge_pos(self, file_folder):
        """write monitoring gauges
        gauges_pos.dat
        file_folder: folder to write file
        gauges_pos: 2-col numpy array of X and Y coordinates
        """
        gauges_pos = self.attributes['gauges_pos']
        file_name = os.path.join(file_folder, 'gauges_pos.dat')
        fmt = ['%g %g']
        fmt = '\n'.join(fmt*gauges_pos.shape[0])
        gauges_pos_str = fmt % tuple(gauges_pos.ravel())
        with open(file_name, 'w') as file2write:
            file2write.write(gauges_pos_str)
        return file_name

    def __write_gauge_ind(self, file_folder):
        """write monitoring gauges index for mult-GPU sections
        gauges_ind.dat
        file_folder: folder to write file
        gauges_ind: 1-col numpy array of index values
        """
        gauges_ind = self.attributes['gauges_ind']
        file_name = os.path.join(file_folder, 'gauges_ind.dat')
        fmt = ['%g']
        fmt = '\n'.join(fmt*gauges_ind.shape[0])
        gauges_ind_str = fmt % tuple(gauges_ind.ravel())
        with open(file_name, 'w') as file2write:
            file2write.write(gauges_ind_str)
        return file_name

    def __copy_to_all_sections(self, file_names):
        """ Copy files that are the same in each sections
        file_names: (str) files written in the first seciton [0]
        boundary source files: h_BC_[N].dat, hU_BC_[N].dat
        rainfall source files: precipitation_source_all.dat
        gauges position file: gauges_pos.dat
        """
        if type(file_names) is not list:
            file_names = [file_names]
        for i in np.arange(1, self.num_of_sections):
            field_dir = self.Sections[i]._data_folders['field']
            for file in file_names:
                shutil.copy2(file, field_dir)
    
    def _setup_by_files(self, data_path):
        """Read files to setup hipims input object
        DEM, landcover, rain_mask, endswith '.gz', '.asc', or '.tif'
        rain_source.csv
        """
        data_path = self.data_path
        ras_file = indep_f._check_raster_exist(os.path.join(data_path, 'DEM'))
        if ras_file is not None:
            self.DEM = Raster(ras_file)
            print(ras_file+ ' read')
        ras_file = indep_f._check_raster_exist(os.path.join(data_path,
                                                            'landcover'))        
        if ras_file is not None:
            self.Landcover = Landcover(ras_file, dem_ras=self.DEM)
            print(ras_file+ ' read')
        
        mask_file = indep_f._check_raster_exist(os.path.join(data_path,
                                                         'rain_mask'))
        if mask_file is None:
            mask_file = 0
        source_file = os.path.join(data_path, 'rain_source.csv')
        if not os.path.isfile(source_file):
            source_file = np.array([[0, 0], [1, 0]])
        self.Rainfall = Rainfall(mask_file, source_file, dem_ras=self.DEM)

#%%****************************************************************************
#************************sub-class definition**********************************
class InputHipimsSub(InputHipims):
    """object for each section, child class of InputHipims

    Attributes:
        sectionNO: the serial number of each section
        _valid_cell_subs: (tuple, int) two numpy array indicating rows and cols
            of valid cells on the local grid
        valid_cell_subsOnGlobal: (tuple, int) two numpy array indicating rows
            and cols of valid cells on the global grid
        shared_cells_id: 2-row shared Cells id on a local grid
        case_folder: input folder of each section
        _outline_cell_subs: (tuple, int) two numpy array indicating rows and 
            cols of valid cells on a local grid
    """
    section_id = 0
    def __init__(self, dem_array, header, case_folder, num_of_sections):
        self.section_id = InputHipimsSub.section_id
        InputHipimsSub.section_id = self.section_id+1
        dem_data = Raster(array=dem_array, header=header)
        super().__init__(dem_data, num_of_sections, case_folder)

#%%****************************************************************************
#********************************Static method*********************************
def load_input_object(filename):
    """load object from a dictionary and return as an InputHipims object
    Args:
        filename: a string giving the object file name
    Return: 
        An object of InputHipims
    """
    obj_dict = indep_f.load_object(filename)
    
    if 'DEM' in obj_dict:
        dem_dict = obj_dict['DEM']
        obj_dem = Raster(array=dem_dict['array'],
                                header=dem_dict['header'])
        obj_dict.pop('DEM')
    else:
        raise ValueError(filename+' has no key: DEM')
    obj_in = InputHipims(dem_data=obj_dem,
                         num_of_sections=obj_dict['num_of_sections'],
                         case_folder=obj_dict['_case_folder'])
    
    if 'Landcover' in obj_dict:
        ld_dict = obj_dict['Landcover']
        mask_header = ld_dict['mask_header']
        mask_dict = ld_dict['mask_dict']
        array_shape = (mask_header['nrows'], mask_header['ncols'])
        mask_array = indep_f._dict2grid(mask_dict, array_shape)
        ras_landcover = Raster(array=mask_array, header=mask_header)
        obj_in.set_landcover(ras_landcover)
        obj_dict.pop('Landcover')
    
    if 'Rainfall' in obj_dict:
        rain_dict = obj_dict['Rainfall']
        mask_header = rain_dict['mask_header']
        mask_dict = rain_dict['mask_dict']
        array_shape = (mask_header['nrows'], mask_header['ncols'])
        mask_array = indep_f._dict2grid(mask_dict, array_shape)
        rain_mask = Raster(array=mask_array, header=mask_header)
        rain_source = np.c_[rain_dict['time_s'], rain_dict['rain_rate']]
        obj_in.set_rainfall(rain_mask, rain_source)
        obj_dict.pop('Rainfall')
    
    bound_dict = obj_dict['Boundary']
    for key, value in bound_dict.items():
        obj_in.Boundary.__dict__[key] = value
    obj_dict.pop('Boundary')
    
    summ_dict = obj_dict['Summary']
    for key, value in summ_dict.items():
        obj_in.Summary.__dict__[key] = value
    obj_dict.pop('Summary')
    
    for key, value in obj_dict.items():
        obj_in.__dict__[key] = value
    
    return obj_in

def main():
    print('Class to setup input data')

if __name__=='__main__':
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author: Xiaodong Ming

"""
Boundary
========

To do:
    * Define boundary conditions for hipims model

-----------

"""
# Created on Tue Mar 31 16:05:27 2020
import warnings
import numpy as np
import pandas as pd
import matplotlib.patches as mplP
from hydro_raster import spatial_analysis as sp
#%% boundary class definition
class Boundary(object):
    """
    Class for boundary conditions

    Attributes:

        num_of_bound: number of boundaries

        type: a list of string 'open', 'rigid', 'fall'
                'open': timeseries of boundary depth/discharge/velocity can 
                    be given for this type. If no given timeseries data,
                    water will flow out flatly
                'rigid': no outlet
                'fall': water flow out like a fall, a fix zero water depth and
                    velocities will be given

        extent: (2-col numpy array) poly points to define the extent of a
                IO boundary. If extent is not given, then the boundary is
                the domain outline

        hSources: a two-col numpy array. The 1st col is time(s). The 2nd
                col is water depth(m)

        hUSources: a two-col numpy array. The 1st col is time(s). The 2nd
                col is discharge(m3/s) or a three-col numpy array, the 2nd
                col and the 3rd col are velocities(m/s) in x and y
                direction, respectively.

        h_code: 3-element int to define the type of depth boundary

        hU_code: 3-element int to define th type of velocity boundary

        description: (str) description of a boundary

    """
    def __init__(self, boundary_list=None, outline_boundary='fall'):
        """Initialise the object

        Args:
            boundary_list: (list of dicts), each dict contain keys (polyPoints,
                type, h, hU) to define a IO boundary's position, type, and
                Input-Output (IO) sources timeseries. Keys including:

                1.polyPoints is a numpy array giving X(1st col) and Y(2nd col)
                    coordinates of points to define the position of a boundary.
                    A bound without polyPoints is regarded as the outline_boundary.

                2.type: string, type of the boundary
                    'open': timeseries of boundary depth/discharge/velocity
                            can be given for this type. If no given timeseries
                            data, water will flow out flatly
                    'rigid': water cannot flow in or out
                    'fall': water flow out like a fall, a fix zero water depth
                            and velocities will be given

                3.h: a two-col numpy array. The 1st col is time(s). The 2nd col
                     is water depth(m)

                4.hU: a two-col numpy array. The 1st col is time(s). The 2nd
                    col is discharge(m3/s) or a three-col numpy array, the 2nd
                    col and the 3rd col are velocities(m/s) in x and y
                    direction, respectively.

            outline_boundary: (str) 'open'|'rigid', default outline boundary is
                open and both h and hU are set as zero
            if h or hU is given, then the boundary type is set as 'open' in 
                function _setup_boundary_data_table
        """
        data_table = _setup_boundary_data_table(boundary_list, outline_boundary)
        data_table = _get_boundary_code(data_table)
        num_of_bound = data_table.shape[0]
        self.data_table = data_table
        self.num_of_bound = num_of_bound
        self.h_sources = data_table['hSources']
        self.hU_sources = data_table['hUSources']
        self.boundary_list = boundary_list
        self.outline_boundary = outline_boundary
        self.cell_subs = None
        self.cell_id = None

    def print_summary(self):
        """Print the summary information
        
        """
        print('Number of boundaries: '+str(self.num_of_bound))
        for n in range(self.num_of_bound):
            if self.cell_subs is not None:
                num_cells = self.cell_subs[n][0].size
                description = self.data_table.description[n] \
                                 + ', number of cells: '+str(num_cells)
                print(str(n)+'. '+description)

    def get_summary(self):
        """ Get summary information strings
        
        """
        summary_dict = {}
        summary_dict['Number of boundaries'] = str(self.num_of_bound)
        summary_str = []
        for n in np.arange(self.num_of_bound):
            if self.cell_subs is not None:
                num_cells = self.cell_subs[n][0].size
                description = self.data_table.description[n] \
                                 + ', number of cells: '+str(num_cells)
                summary_str.append(str(n)+'. '+description)
        summary_dict['Boundary details'] = summary_str
        return summary_dict

    def _fetch_boundary_cells(self, valid_subs, outline_subs, dem_header):
        """ To get the subsripts and id of boundary cells on the domain grid
        
        valid_subs, outline_subs, dem_header are from hipims object
            _valid_cell_subs, _outline_cell_subs
        cell_subs: (tuple)subscripts of outline boundary cells
        cell_id: (numpy vector)valid id of outline boundary cells
        """
        # to get outline cell id based on _outline_cell_subs
        vector_id = np.arange(valid_subs[0].size)
        nrows = dem_header['nrows']
        ncols = dem_header['ncols']
        cellsize = dem_header['cellsize']
        xllcorner = dem_header['xllcorner']
        yllcorner = dem_header['yllcorner']
        grid_cell_id = np.zeros((nrows, ncols))
        grid_cell_id[valid_subs] = vector_id
        outline_id = grid_cell_id[outline_subs]
        outline_id = outline_id.astype('int64')
        # to get boundary cells based on the spatial extent of each boundary
        bound_cell_x = xllcorner+(outline_subs[1]+0.5)*cellsize
        bound_cell_y = yllcorner+(nrows-outline_subs[0]-0.5) *cellsize
        n = 1 # sequence number of boundaries
        data_table = self.data_table
        cell_subs = []
        cell_id = []
        for n in range(data_table.shape[0]):
            if data_table.extent[n] is None: #outline boundary
                dem_extent = sp.header2extent(dem_header)
                polyPoints = sp.extent2shape_points(dem_extent)
            elif len(data_table.extent[n]) == 2:
                xyv = data_table.extent[n]
                polyPoints = sp.extent2shape_points([np.min(xyv[:, 0]),
                                                     np.max(xyv[:, 0]),
                                                     np.min(xyv[:, 1]),
                                                     np.max(xyv[:, 1])])
            else:
                polyPoints = data_table.extent[n]
            poly = mplP.Polygon(polyPoints, closed=True)
            bound_cell_xy = np.array([bound_cell_x, bound_cell_y])
            bound_cell_xy = np.transpose(bound_cell_xy)
            ind1 = poly.contains_points(bound_cell_xy)
            row = outline_subs[0][ind1]
            col = outline_subs[1][ind1]
            cell_id.append(outline_id[ind1])
            cell_subs.append((row, col))
        self.cell_subs = cell_subs
        self.cell_id = cell_id
        self.cell_subs_wet_io = _find_wet_io_cells(self, source_key='hUSources')
    
    def _convert_flow2velocity(self, dem_obj):
        """ Convert 2-col flow timeseries to 3-col velocity timeseries
        in datatable
        """
        hU_sources = self.data_table['hUSources']
        for i in np.arange(self.num_of_bound):
            hU_source = hU_sources[i]
            cell_subs = self.cell_subs[i] # row and col
            if hU_source is not None:
                if hU_source.shape[1] == 2:
                    if cell_subs[0].size == 1:
                        warnings.warn('Only one cell on boundary '+str(i)+
                              ', you should better convert flow to velocities by yourself')
                # flow is given, no velocity
                    theta = _get_bound_normal(cell_subs, dem_obj)
                    boundary_length = cell_subs[0].size*dem_obj.header['cellsize']
                    hUx = hU_source[:, 1]*np.cos(theta)/boundary_length
                    hUy = hU_source[:, 1]*np.sin(theta)/boundary_length
                    hU_source = np.c_[hU_source[:, 0], hUx, hUy]
                    self.data_table['hUSources'][i] = hU_source
                    print('Flow series on boundary '+str(i)+
                              ' is converted to velocities')
                    print('Theta = '+'{:.2f}'.format(theta/np.pi*180)+'degree')
        self.hU_sources = self.data_table['hUSources']

    def _divide_domain(self, hipims_obj):
        """ Create Boundary objects for each sub-domain
        
        IF hipims_obj has sub sections
        """
        boundary_list = hipims_obj.Boundary.boundary_list
        outline_boundary = hipims_obj.Boundary.outline_boundary
        header_global = hipims_obj.header
        outline_subs = hipims_obj._outline_cell_subs
        for i in range(hipims_obj.num_of_sections):
            obj_section = hipims_obj.Sections[i]
            header_local = obj_section.header
            # convert global subscripts to local
            outline_subs_local = _cell_subs_convertor(
                outline_subs, header_global, header_local, to_global=False)
            valid_subs_local = obj_section._valid_cell_subs
            bound_obj = Boundary(boundary_list, outline_boundary)
            bound_obj._fetch_boundary_cells(
                valid_subs_local, outline_subs_local, header_local)
            obj_section.Boundary = bound_obj
#            summary_str = bound_obj.get_summary()
#            obj_section.Summary.add_items('Boundary conditions', summary_str)

#%%
# private function called by Class Boundary
def _setup_boundary_data_table(boundary_list, outline_boundary='open'):
    """ Initialize boundary data table based on boundary_list
    Add attributes type, extent, hSources, hUSources
    boundary_list: (list) of dict with keys polyPoints, h, hU
    outline_boundary: (str) 'open', 'rigid', 'fall'
    A bound without polyPoints is regarded as the outline_boundary
    If h or hU is given, then the boundary type is set as 'open', which means
     'h' and 'hU' have priority to 'type'
    """
    data_table = pd.DataFrame(columns=['type', 'extent',
                                       'hSources', 'hUSources',
                                       'h_code', 'hU_code', 'name'])
    # set default outline boundary [0]
    if outline_boundary == 'fall':
        hSources = np.array([[0, 0], [1, 0]])
        hUSources = np.array([[0, 0, 0], [1, 0, 0]])
        new_row = pd.DataFrame({'type':['fall'], 
                                'extent':[None],
                                'hSources':[hSources],
                                'hUSources':[hUSources]})
    elif outline_boundary == 'rigid':
        new_row = pd.DataFrame({'type':['rigid'], 
                                'extent':[None],
                                'hSources':[None], 
                                'hUSources':[None]})
    elif outline_boundary == 'open':
        new_row = pd.DataFrame({'type':['open'], 
                                'extent':[None],
                                'hSources':[None],
                                'hUSources':[None]})
    else:
        raise ValueError("outline_boundary must be fall, open or rigid!")
    data_table = pd.concat([data_table, new_row], ignore_index=True)
    data_table.name[0] = 'Outline boundary'
    # convert boundary_list to a dataframe
    if boundary_list is None:
        boundary_list = []
    for one_bound in boundary_list:
        # Only a bound with polyPoints can be regarded as a boundary
        # Otherwise it will be treated as the outline boundary
        bound_ind = data_table.shape[0]  # bound index
            # set extent
        if 'polyPoints' in one_bound.keys():
            bound_extent = np.array(one_bound['polyPoints'])
            new_row = pd.DataFrame({'extent':[bound_extent]})
            data_table = pd.concat([data_table, new_row], 
                                    ignore_index=True)
        else:
            bound_ind = 0
            print('polyPoints is not given in boundary_list[0],'
                  'set as outline boundary')
        if 'type' in one_bound.keys():
            bound_type = one_bound['type']
        else:
            bound_type = 'open' # defaul type is open

        if 'h' in one_bound.keys():
            data_table.hSources[bound_ind] = np.array(one_bound['h'])
            bound_type = 'open'
        else:
            data_table.hSources[bound_ind] = None
        
        if 'hU' in one_bound.keys():
            data_table.hUSources[bound_ind] = np.array(one_bound['hU'])
            bound_type = 'open'
        else:
            data_table.hUSources[bound_ind] = None
        
        data_table.type[bound_ind] = bound_type
        if 'name' in one_bound.keys():
            data_table.name[bound_ind] = one_bound['name']

    return data_table

# private function called by Class Boundary
def _get_boundary_code(boudnary_data_table):
    """ Get the 3-element boundary code for h and hU
    boudnary_data_table: the boundary data table with
        columns ['type', 'hSources', 'hUSources']
    Return a new data_table added with h_code, hU_code, and description
    """
#        Get the three column boundary code
        #default outline boundary
    data_table = boudnary_data_table
    num_of_bound = data_table.shape[0]
    description = []
    n_seq = 0  # sequence of boundary
    m_h = 0  # sequence of boundary with IO source files
    m_hu = 0
    for n_seq in range(num_of_bound):
        
        data_table.h_code[n_seq] = np.array([[2, 0, 0]])
        bound_type = data_table.type[n_seq]
        if bound_type == 'rigid':
            data_table.hU_code[n_seq] = np.array([[2, 2, 0]])
            description1 = bound_type
        elif bound_type == 'fall':
            h_sources = np.array([[0, 0], [1, 0]])
            hU_sources = np.array([[0, 0, 0], [1, 0, 0]])
            data_table.h_code[n_seq] = np.array([[3, 0, m_h]])
            data_table.hU_code[n_seq] = np.array([[3, 0, m_hu]])
            m_h = m_h+1
            m_hu = m_hu+1
            description1 = bound_type+', h and hU fixed as zero'
        else: # open
            h_sources = data_table.hSources[n_seq]
            hU_sources = data_table.hUSources[n_seq]
            data_table.hU_code[n_seq] = np.array([[2, 1, 0]])
            description1 = bound_type
            if h_sources is not None:
                data_table.h_code[n_seq] = np.array([[3, 0, m_h]]) #[3 0 m]
                description1 = description1+', h given'
                m_h = m_h+1
            if hU_sources is not None:
                data_table.hU_code[n_seq] = np.array([[3, 0, m_hu]]) #[3 0 m]
                description1 = description1+', hU given'
                m_hu = m_hu+1
        description.append(description1)
    description[0] = '(outline) '+ description[0] # indicate outline boundary
    data_table['description'] = description
    return data_table

def _cell_subs_convertor(input_cell_subs, header_global,
                         header_local, to_global=True):
    """
    Convert global cell subs to divided local cell subs or the otherwise
    and return output_cell_subs, only rows need to be changed
    input_cell_subs : (tuple) input rows and cols of a grid
    header_global : head information of the global grid
    header_local : head information of the local grid
    to_global : logical values, True (local to global) or
                                False(global to local)
    Return:
        output_cell_subs: (tuple) output rows and cols of a grid
    """
    # X and Y coordinates of the centre of the first cell
    y00_centre_global = header_global['yllcorner']+\
                         (header_global['nrows']+0.5)*header_global['cellsize']
    y00_centre_local = header_local['yllcorner']+\
                        (header_local['nrows']+0.5)*header_local['cellsize']
    row_gap = (y00_centre_global-y00_centre_local)/header_local['cellsize']
    row_gap = round(row_gap)
    rows = input_cell_subs[0]
    cols = input_cell_subs[1]
    if to_global:
        rows = rows+row_gap
        # remove subs out of range of the global DEM
        ind = np.logical_and(rows >= 0, rows < header_global['nrows'])
    else:
        rows = rows-row_gap
        # remove subs out of range of the global DEM
        ind = np.logical_and(rows >= 0, rows < header_local['nrows'])
    rows = rows.astype(cols.dtype)
    rows = rows[ind]
    cols = cols[ind]
    output_cell_subs = (rows, cols)
    return output_cell_subs

def _find_wet_io_cells(bound_obj, source_key='hUSources'):
    """ return subs of wet IO boundary cells
    """
    wet_cell_rows = []
    wet_cell_cols = []
    for ind_num in np.arange(bound_obj.num_of_bound):
        bound_source = bound_obj.data_table[source_key][ind_num]
        if bound_source is not None:
            source_value = np.unique(bound_source[:, 1:])
            # zero boundary conditions
            if not (source_value.size == 1 and source_value[0] == 0):
                cell_subs = bound_obj.cell_subs[ind_num]
                wet_cell_rows.append(cell_subs[0])
                wet_cell_cols.append(cell_subs[1])
    if len(wet_cell_rows)>0:
        wet_cell_rows = np.concatenate(wet_cell_rows, axis=0 )
        wet_cell_cols = np.concatenate(wet_cell_cols, axis=0 )
        cell_subs_wet_io = (wet_cell_rows, wet_cell_cols)
    else:
        cell_subs_wet_io = None           
    return cell_subs_wet_io

def _get_bound_normal(cell_subs, dem_obj):
    """get the angle of the normal vector of the bound line
    """
    cell_xy = sp.sub2map(cell_subs[0], cell_subs[1], dem_obj.header)
    cell_xy = np.c_[cell_xy[0], cell_xy[1]]
    if np.unique(cell_subs[1]).size==1: # vertical bound line
        theta = 0
    elif np.unique(cell_subs[0]).size==1: # horizontal bound line
        theta = np.pi*0.5
    else:
        boundary_slope = np.polyfit(cell_xy[:, 0], cell_xy[:, 1], 1)
        theta = np.arctan(boundary_slope[0])+np.pi/2
        if theta > np.pi*2:
            theta = theta-np.pi*2
        if theta < 0:
            theta = theta+np.pi*2
    start_xy = cell_xy[0]
    unit_vec = np.array([np.cos(theta), np.sin(theta)])
    end0_xy =  unit_vec*4*dem_obj.header['cellsize']+start_xy
    end1_xy = -unit_vec*4*dem_obj.header['cellsize']+start_xy
    end0_value = _get_array_value_by_rc(end0_xy, dem_obj)
    end1_value = _get_array_value_by_rc(end1_xy, dem_obj)
    if np.isnan(end0_value) & ~np.isnan(end1_value): # towards out
    # reverse direction of the boundline normal
        if theta >= np.pi:
            theta = theta-np.pi
        else:
            theta = theta+np.pi
    elif ~np.isnan(end0_value) & np.isnan(end1_value): # towards in
        pass
    else:
        warnings.warn('Cannot judge flow direction, please double check your boundary condition')
    return theta

def _get_array_value_by_rc(end_xy, dem_obj):
    rc_num = sp.map2sub(end_xy[0], end_xy[1], dem_obj.header)
    try:
        end_value = dem_obj.array[rc_num]
    except:
        end_value = np.nan
    if end_value == dem_obj.header['NODATA_value']:
        end_value == np.nan
    return end_value


# -*- coding: utf-8 -*-
# @Author: Xiaodong Ming
# @Date:   2021-09-02 11:34:09
# @Last Modified by:   Xiaodong Ming
# @Last Modified time: 2022-03-22 18:19:42
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Fri Jun  5 15:44:18 2020
#Author: Xiaodong Ming

"""
Summary
=======

To do:
    To generate, store, write, and read summary information of a flood model

-----------------

"""
#%%
import os
import copy
import numpy as np
import json
from . import rainfall_processing as rp
class Summary:
    """ Summary information of a flood case

    Attributes:

        grid_attr:
        model_attr:
        boundary_attr:
        rain_attr:
        params_attr:
        initial_attr:
        
    """
    # default parameters
    grid_attr = {'area':0, # domain area in meter
                 'shape':(1, 1), # tuple of int, grid size rows and cols
                 'cellsize':1,
                 'num_cells':0, # number of valid cells
                 'extent_dict':{'left':0, 'right':0, 'bottom':0, 'top':0},
                 }
    model_attr = {'case_folder':None, # string
                  'birthday':None,
                  'num_GPU':1,
                  'run_time':(0, 3600, 600, 3600), # start, end, write_interval
                  'num_gauges': 0} 
                              
    boundary_attr = {'num_boundary':1,
                     'boundary_details':'outline'}
    rain_attr = {'num_source':0,
                 'average':0,
                 'sum':0,
                 'max':0,
                 'spatial_res':1000, # meter
                 'temporal_res':0 # second
                 }
    params_attr = {'manning':0, 'sewer_sink':0,
                   'cumulative_depth':0, 'hydraulic_conductivity':0,
                   'capillary_head':0, 'water_content_diff':0}
    initial_attr = {'h0':0, 'hU0x':0, 'hU0y':0}
    
    
    def __init__(self, data_in): 
        if type(data_in) is str: # a jason file
            self.__setup_from_file(data_in)
        elif type(data_in) is dict: # a jason file
            for key, value in data_in.items():
                self.__dict__[key] = value
        elif hasattr(data_in, 'DEM'): # a case_obj or dem_obj
            self.__setup_from_object(data_in)

    def __str__(self):
        """
        To show object summary information when it is called in console
        """
        self.display()
        return  self.__class__.__name__
#    __repr__ = __str__
    
    def __setup_from_file(self, filename):
        """Setup object from a file
        """
        summary_dict = load_summary(filename)
        self.grid_attr = summary_dict['grid_attr']
        self.model_attr = summary_dict['model_attr']
        self.boundary_attr = summary_dict['boundary_attr']
        self.rain_attr = summary_dict['rain_attr']
        self.initial_attr = summary_dict['initial_attr']
        self.params_attr = summary_dict['params_attr']
    
    def __setup_from_object(self, obj_in):
        """Setup from a InputHipims object
        """
        
        dem_array = obj_in.DEM.array
        dem_header = obj_in.DEM.header
        extent_dict = obj_in.DEM.extent_dict
        self.set_grid_attr(dem_array, dem_header, extent_dict)

        dt_fmt = '%Y-%m-%d %H:%M'
        self.set_model_attr(case_folder=obj_in.get_case_folder(),
                            birthday=obj_in.birthday.strftime(dt_fmt),
                            num_GPU=obj_in.num_of_sections,
                            run_time=[obj_in.times.tolist(), 's'],
                            num_gauges=obj_in.attributes['gauges_pos'].shape[0]
                            )
        if hasattr(obj_in, 'Rainfall'):
            self.rain_attr = obj_in.Rainfall.attrs
        else:
            rain_source = obj_in.attributes['precipitation_source']
            rain_mask = obj_in.attributes['precipitation_mask']
            rain_mask = rain_mask+dem_array*0
            cellsize = obj_in.DEM.header['cellsize']
            self.set_rain_attr(rain_source, rain_mask, cellsize)
        
        if hasattr(obj_in, 'Boundary'):
            self.set_boundary_attr(obj_in.Boundary)
        
        self.valid_ind = ~np.isnan(dem_array)
        args_dict = copy.copy(self.params_attr)
        for param_name in args_dict.keys():
            args_dict[param_name] = obj_in.attributes[param_name]
        self.set_params_attr(**args_dict)
        self.set_initial_attr(h0=obj_in.attributes['h0'],
                              hU0x=obj_in.attributes['hU0x'],
                              hU0y=obj_in.attributes['hU0y'])

    def set_boundary_attr(self, obj_boundary):
        self.boundary_attr['num_boundary'] = obj_boundary.num_of_bound
        smry_dict = obj_boundary.get_summary()
        self.boundary_attr['boundary_details'] = smry_dict['Boundary details']
        
    def set_params_attr(self, **kw):
        names = self.params_attr.keys()
        for key, value in kw.items():
            if key in names:
                if np.array(value).size==1:
                    self.params_attr[key] = value
                else:
                    values = value[self.valid_ind]
                    values, counts = np.unique(values, return_counts=True)
                    ratios = counts/counts.sum()
                    ratios = ratios.round(4)*100
                    ratios_str = str(ratios.tolist())+'%'
                    if key in ['sewer_sink', 'hydraulic_conductivity']:
                        values = np.round(values*3600*1000, 4)
                        ratios_str = ratios_str+' (unit: mm/h)'
                    self.params_attr[key] = [values.tolist(), ratios_str]
            else:
                print(key+' is not a grid parameter')
    
    def set_initial_attr(self, **kw):
        names = self.initial_attr.keys()
        for key, value in kw.items():
            if key in names:
                if np.array(value).size==1:
                    self.initial_attr[key] = value
                else:
                    values = value[self.valid_ind]
                    ratios = np.sum(values != 0)/values.size
                    self.initial_attr[key] = 'Wet ratio: {:.2f}%'.format(
                            ratios*100)
            else:
                print(key+' is not a grid parameter')
    
    def set_grid_attr(self, dem_array, dem_header, extent_dict):
        """ assign grid_attr
        """
        cellsize = dem_header['cellsize']
        shape = dem_array.shape
        num_cells = ~np.isnan(dem_array)
        num_cells = num_cells.sum()
        area_value = num_cells*cellsize**2
        if area_value>1e7:
            area_value = area_value/1e6
            area_unit = 'km^2'
        else:
            area_unit = 'm^2'
        grid_attr = {'area':[area_value, area_unit],
                     'shape':shape,
                     'cellsize':[cellsize, 'm'],
                     'num_cells':int(num_cells),
                     'extent':extent_dict
                     }
        self.grid_attr = grid_attr  
        
    def set_model_attr(self, **kw):
        """ assign case_folder, num_GPU, run_time, gauge_pos
        """
        for key, value in kw.items():
            if key=='gauges_pos':
                value = np.array(value)
                num_gauges=value.shape[0]
                self.model_attr['num_gauges'] = num_gauges
            else:
                if type(value) is np.ndarray:
                    value = value.tolist()            
                self.model_attr[key] = value
    
    def set_rain_attr(self, rain_source, rain_mask, cellsize):
        """ define rain_attr
        """
        times = rain_source[:,0]
        temporal_res = (times.max()-times.min())/times.size # seconds
        num_cells = rain_mask[~np.isnan(rain_mask)]
        num_source = np.unique(num_cells).size
        spatial_res = np.sqrt(num_cells.size*cellsize/num_source)
        data_sum = rp.get_time_series(rain_source, rain_mask, method='sum')
        rain_total = np.trapz(data_sum[:,1], x=times, axis=0)/1000 # mm
        rain_mean = rain_total/(times.max()-times.min())*3600 # mm/h
        data_max = rp.get_time_series(rain_source, rain_mask, method='max')
        self.rain_attr['num_source'] = num_source
        self.rain_attr['max'] = [np.max(data_max[:,1]).round(2), 'mm/h']
        self.rain_attr['sum'] = [rain_total.round(2), 'mm']
        self.rain_attr['average'] = [rain_mean.round(2), 'mm/h']
        self.rain_attr['spatial_res'] = [spatial_res.round(), 'm']
        self.rain_attr['temporal_res'] = [temporal_res.round(), 's']
    
    def to_dict(self):
        """Convert the object to a dictionary
        """        
        def set_top_dict(origin_dict):
            top_dict = {}    
            for key, value in origin_dict.items():
                if (type(value) is list) or (type(value) is tuple):
                    top_dict[key] = str(value)
                else:
                    top_dict[key] = value
            return top_dict

        summary_dict = {
                'grid_attr':set_top_dict(self.grid_attr),
                'model_attr':set_top_dict(self.model_attr),
                'initial_attr': set_top_dict(self.initial_attr),
                'boundary_attr':set_top_dict(self.boundary_attr),
                'rain_attr':set_top_dict(self.rain_attr),
                'params_attr':set_top_dict(self.params_attr)
                }
        return summary_dict

    def display(self):
        """ display summary information
        """
        def print_dict(one_dict):
            for key, value in one_dict.items():
                print(key,':',value)
        print('---------------------- Model information ---------------------')
        print_dict(self.model_attr)
        print('---------------------- Grid information ----------------------')
        print_dict(self.grid_attr)
        print('---------------------- Initial condition ---------------------')
        print_dict(self.initial_attr)
        print('---------------------- Boundary condition --------------------')
        print_dict(self.boundary_attr)
        print('---------------------- Rainfall ------------------------------')
        print_dict(self.rain_attr)
        print('---------------------- Parameters ----------------------------')
        print_dict(self.params_attr)
        
    def to_json(self, file_name=None):
        """ Write to json file
        """
        summary_dict = self.to_dict()
        if file_name is None:
            case_folder = self.model_attr['case_folder']
            file_name = os.path.join(case_folder, 'readme.txt')
        with open(file_name, 'w') as fp:
            json.dump(summary_dict, fp, indent=4)

    def write_readme(self, filename=None):
        self.to_json(filename)

def load_summary(file_name):
    """ Read a json text file as a dict of Summary
    """
    def js_r(filename):
        with open(filename) as f_in:
            return(json.load(f_in))
    summary_dict = js_r(file_name)
    for key in summary_dict.keys():
        for key_1, value_1 in summary_dict[key].items():
            if type(value_1) is str:
                if value_1[0] == '[' or value_1[0] == '(':
                    summary_dict[key][key_1] = eval(value_1)
    return summary_dict    

def main():
    print('Class to show intput Summary')

if __name__=='__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Tue Jun 23 21:32:54 2020
# Author: Xiaodong Ming

"""
Rainfall
========

To do:
    To read, compute, and show rainfall data

-----------------

"""
import time
import numpy as np
import matplotlib.pyplot as plt	
import matplotlib.dates as mdates
from scipy import stats
from datetime import datetime
from datetime import timedelta
from hydro_raster.Raster import Raster
from . import indep_functions as indep_f
class Rainfall:
    """ a class to set rainfall data for hipims

    Attributes:

        time_s: 1-col array, time in seconds
        mask_header: dictionary showing mask georeference
        mask_dict: dict with two keys:'value' and 'index', providing int array
                showing rain source number and their index respectively
        rain_rate: numpy array m/s
        attrs: summary of the object
        subs_in: tuple of row and col number, provide subs of the mask array values
            inside the model domain. Only available when dem_ras is given
        start_date: start date and time of time zero in time_s
        time_dt: datetime format of time_s

    """
    def __init__(self, rain_mask, rain_source, source_sep=',', dem_ras=None):
        """initialize rainfall object with source file and mask file

        Args:
            rain_mask: str [filename of a Raster endswith .gz/asc/tif] numpy int array with the same shape with DEM array
                 a Raster object. If rain_mask is a scalar or array, dem_ras must be provided.           
            rain_source: numpy array the 1st column is time in seconds, 2nd to the end columns are rainfall rates in m/s. 
                str [filename of a csv file for rainfall source data]
            dem_ras: a Raster object for DEM
            source_sep: delimeter of the rain source file
            
        """
        self.set_mask(rain_mask, dem_ras)
        self.set_source(rain_source, source_sep)
        del rain_mask
        
    def set_mask(self, rain_mask, dem_ras=None):
        """Set rainfall mask from a scalar or a grid (object/file)

        if rain_mask is a scalar or array, dem_ras must be provided
        """
        if type(rain_mask) is str:
            obj_rain_mask = Raster(rain_mask)
            print(rain_mask+ ' read')
        elif hasattr(rain_mask, 'header'):
            obj_rain_mask = rain_mask
        else: # numpy array or a scalar, dem_ras must be provided
            rain_mask = np.array(rain_mask).astype('int32')
            dem_shape = dem_ras.array.shape
            if rain_mask.size > 1:
                if rain_mask.shape != dem_shape:
                    raise ValueError('The shape of rainfall_mask array '
                             'is not consistent with DEM')
            mask_array = np.zeros(dem_shape, dtype=np.int32)+rain_mask
            obj_rain_mask = Raster(array=mask_array, header=dem_ras.header)
        if hasattr(dem_ras, 'header'):
            # rain mask resample to the same shape with DEM
            self.mask_dict = indep_f._mask2dict(obj_rain_mask, dem_ras.header)
            self.mask_header = dem_ras.header
            self.subs_in = np.where(~np.isnan(dem_ras.array))
        else:
            self.mask_dict = indep_f._mask2dict(obj_rain_mask)
            self.mask_header = obj_rain_mask.header
    
    def set_source(self, rain_source, delimiter=','):
        """Set rainfall mask from a numpy array or a csv file
        """
        if type(rain_source) is str:
            print('reading '+rain_source)
            rain_source = np.loadtxt(rain_source, delimiter=delimiter)
        elif type(rain_source) is np.ndarray:
            pass
        else:
            raise IOError('rain_source must be either a filename or a numpy'
                          ' array')
#        _ = rp._check_rainfall_rate_values(rain_source, times_in_1st_col=True)
        self.time_s = rain_source[:, 0 ]
        self.rain_rate = rain_source[:, 1:]        
        self.get_attrs()

    def get_time_series(self, method='mean', rain_rate_valid=None):
        """ Plot time series of average rainfall rate inside the model domain   

        Args:
            method: 'mean'|'max','min','mean'method to calculate gridded rainfall 
                over the model domain
        """
        start = time.perf_counter()
        if rain_rate_valid is None:
            rain_rate_valid, _ = self.get_valid_rain_rate() # time-source id
        if method == 'mean':
            value_y = np.mean(rain_rate_valid, axis=1) # m/s
        elif method == 'max':
            value_y = np.max(rain_rate_valid, axis=1) # m/s
        elif method == 'min':
            value_y = np.min(rain_rate_valid, axis=1) # m/s
        elif method == 'median':
            value_y = np.median(rain_rate_valid, axis=1) # m/s
        elif method == 'sum':
            value_y = np.sum(rain_rate_valid, axis=1) # m/s
        else:
            raise ValueError('Cannot recognise the calculation method')
        value_y =  value_y*3600*1000 # mm/h
        time_series = np.c_[self.time_s, value_y]
        end = time.perf_counter()
        print('get_time_series took'+str(end-start))
        return time_series

    def get_spatial_map(self, method='sum'):
        """Get spatial rainfall map over time series

        Args:
            rain_mask_obj: asc file name or Raster object for rain mask
            cellsize: resample the rain_mask to a new grid (with larger cellsize)
            method: sum|mean caculate method for each cell, sum by time or mean by time 

        """
        # caculate rain source
        time_s = self.time_s
        rain_total = np.trapz(self.rain_rate, time_s, axis=0)*1000 #mm
        unit_str = 'mm/h'
        if method == 'sum':
            cell_rain = rain_total #mm
            unit_str = 'mm'
        elif method == 'mean':
            cell_rain = rain_total/(time_s.max()-time_s.min())*3600 #mm/h
        elif method == 'max':
            cell_rain = np.max(self.rain_rate, axis=0)*1000*3600 #mm/h
        elif method== 'min':
            cell_rain = np.min(self.rain_rate, axis=0)*1000*3600 #mm/h
        elif method== 'median':
            cell_rain = np.median(self.rain_rate, axis=0)*1000*3600 #mm/h
        else:
            raise ValueError('Cannot recognise the calculation method')    
        mask_values = self.mask_dict['value']
        rain_values = cell_rain[mask_values.astype('int32')]
        rain_dict = {'value':rain_values, 'index':self.mask_dict['index']}
        array_shape = (self.mask_header['nrows'], self.mask_header['ncols'])
        rain_array = indep_f._dict2grid(rain_dict, array_shape)
        return rain_array, unit_str
    
    def get_valid_rain_rate(self, unique=True):
        """Get a rain rate array over valid cells

        row is time axis, col is source id axis

        """
        array_shape = (self.mask_header['nrows'], self.mask_header['ncols'])
        mask_array = indep_f._dict2grid(self.mask_dict, array_shape)
        
        if hasattr(self, 'subs_in'):
            mask_valid = mask_array[self.subs_in]
        else:
            mask_valid = mask_array.flatten()
        mask_valid = mask_valid.astype('int32')
        if unique:
            rain_rate_valid = self.rain_rate[:, np.unique(mask_valid)]
        else:
            rain_rate_valid = self.rain_rate[:, mask_valid]
        return rain_rate_valid, mask_valid
    
    def set_start_date(self, datetime_str, fmt):
        """Set start date for the time series
        """
        self.start_date = datetime.strptime(datetime_str, fmt)
        time_delta = np.array([timedelta(seconds=i) for i in self.time_s])
        self.time_dt = self.start_date+time_delta
    
    def get_attrs(self):
        """ define rain_attr
        """
        rain_rate_valid, mask_valid = self.get_valid_rain_rate()
        time_s = self.time_s
        cellsize = self.mask_header['cellsize']
        temporal_res = (time_s.max()-time_s.min())/(time_s.size-1) # seconds
        source_uniq, source_counts = np.unique(mask_valid, return_counts=True)
        num_source = source_uniq.size
        mode_value = stats.mode(source_counts)[0]
        mode_value = np.array(mode_value).flatten()
        spatial_res = np.sqrt(mode_value[0])*cellsize
        cell_mean = np.mean(rain_rate_valid, axis=1)
        # total value along time axis for a cell
        rain_total = np.trapz(cell_mean, x=time_s, axis=0)*1000 # mm
        rain_mean = rain_total/(time_s.max()-time_s.min())*3600 # mm/h
        cell_max = np.max(rain_rate_valid, axis=1)*3600*1000
        attrs = {}
        attrs['num_source'] = num_source
        attrs['max'] = [np.max(cell_max).round(2), 'mm/h']
        attrs['sum'] = [rain_total.round(2), 'mm']
        attrs['average'] = [rain_mean.round(2), 'mm/h']
        attrs['spatial_res'] = [spatial_res.round(), 'm']
        attrs['temporal_res'] = [temporal_res.round(), 's']
        self.attrs = attrs
        return attrs
    
    def get_source_array(self):
        """Return a source array, first column is time_s, then continue with
        rain_rate in m/s
        """
        rain_source = np.c_[self.time_s, self.rain_rate]
        return rain_source
    
    def get_mask_array(self):
        """Return a mask array
        """
        array_shape = (self.mask_header['nrows'], self.mask_header['ncols'])
        mask_array = indep_f._dict2grid(self.mask_dict, array_shape)
        return mask_array
    
#%% Visualization	
    def plot_time_series(self, method='mean', dt_interval=24, 	
                         dt_format='%m-%d', title_str=None,	
                         **kwargs):	
        """ Plot time series of average rainfall rate inside the model domain   	
        method: 'mean'|'max','min','mean'method to calculate gridded rainfall 	
        over the model domain	
        """	
        value_y = self.get_time_series(method=method)
        value_y = value_y[:, -1]
        fig, ax = plt.subplots()        	
        if hasattr(self, 'time_dt'):	
            time_x = self.time_dt	
            ax.xaxis.set_major_locator(	
                    mdates.HourLocator(interval=dt_interval))	
            ax.xaxis.set_major_formatter(mdates.DateFormatter(dt_format))	
        else:	
            time_x = self.time_s	
        ax.plot(time_x, value_y, **kwargs)	
        ax.set_ylabel(method+' rainfall rate (mm/h)')	
        ax.grid(True)	
        if title_str is None:	
            title_str = method+' precipitation in the model domain'	
        ax.set_title(title_str)	
        plt.show()	
        return fig, ax

    def plot_rainfall_map(self, method='sum',cmap='YlGnBu', **kw):	
        """plot rainfall map within model domain	
        method: the way to calculate time series rainfall rate	
        """	
        rain_array, unit_str = self.get_spatial_map(method)	
        grid_obj = Raster(array=rain_array, header=self.mask_header)	
        fig, ax = grid_obj.mapshow(cax_str=unit_str, cmap=cmap, **kw)	
        return fig, ax	



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Dec  4 14:26:04 2019
# Author: Xiaodong Ming

"""
rainfall_processing
===================

To do:
    Processing rainfall


Explaination of general parameters

    rain_source: (numpy array) rainfall source array
        The 1st column is usually time series in seconds, from the 2nd column towards end columns are rainfall rate in m/s

    rain_mask: (numpy array) provide sequnce number of each gridded rainfall source

    start_date: a datetime object to give the initial date and time of rain



----------------

"""
import os
import warnings
import imageio
import shapefile # Requires the pyshp package
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from datetime import datetime
from datetime import timedelta
from hydro_raster.Raster import Raster
#%%
def get_time_series(rain_source, rain_mask=None, 
                     start_date=None, method='mean'):
    """ Plot time series of average rainfall rate inside the model domain

    Args:
        method: 'mean'|'max','min','mean'method to calculate gridded rainfall over the model domain

    """
    if rain_mask is not None:
        rain_mask = rain_mask[~np.isnan(rain_mask)]
        rain_mask = rain_mask.astype('int32')
        rain_mask_unique = np.unique(rain_mask).flatten()
        rain_mask_unique = rain_mask_unique.astype('int32') 
        rain_source_valid = rain_source[:, rain_mask_unique+1]
    else:
        rain_source_valid = rain_source[:, 1:]
    time_series = rain_source[:,0]
    if type(start_date) is datetime:
        time_delta = np.array([timedelta(seconds=i) for i in time_series])
        time_x = start_date+time_delta
    else:
        time_x = time_series
    if method == 'mean':
        value_y = np.mean(rain_source_valid, axis=1)
    elif method == 'max':
        value_y = np.max(rain_source_valid, axis=1)
    elif method == 'min':
        value_y = np.min(rain_source_valid, axis=1)
    elif method == 'median':
        value_y = np.median(rain_source_valid, axis=1)
    elif method == 'sum':
        value_y = np.sum(rain_source_valid, axis=1)
    else:
        raise ValueError('Cannot recognise the calculation method')
    value_y =  value_y*3600*1000
    plot_data = np.c_[time_x, value_y]
    return plot_data

def get_spatial_map(rain_source, rain_mask_obj, figname=None,
                    method='sum', cellsize=None,
                    shp_file=None, dpi=200, title=None, **kwargs):
    """Get spatial rainfall map

    Args:
        rain_mask_obj: asc file name or Raster object for rain mask
        cellsize: resample the rain_mask to a new grid (with larger cellsize)
        method: sum|mean caculate method for each cell, sum by time or mean by time

    """
    # caculate rain source
    times = rain_source[:,0]
    rain_values = rain_source[:,1:]
    cax_str = 'mm/h'
    if rain_values.shape[0] == 1:
        rain_array = rain_values*1000*3600
        rain_array = np.reshape(rain_array, rain_mask_obj.shape)
        rain_array[rain_array==0] = np.nan
        rain_map_obj = Raster(array=rain_array, header=rain_mask_obj.header)
    else:
        rain_total = np.trapz(rain_values, x=times, axis=0)*1000 #mm
        if method == 'sum':
            cell_rain = rain_total #mm
            cax_str = 'mm'
        elif method == 'mean':
            cell_rain = rain_total/(times.max()-times.min())*3600 #mm/h
        elif method == 'max':
            cell_rain = np.max(rain_values, axis=0)*1000*3600 #mm/h
        elif method== 'min':
            cell_rain = np.min(rain_values, axis=0)*1000*3600 #mm/h
        elif method== 'median':
            cell_rain = np.median(rain_values, axis=0)*1000*3600 #mm/h
        else:
            raise ValueError('Cannot recognise the calculation method')    
        # get spatial data
        if type(rain_mask_obj) is str:
            rain_mask_obj = Raster(rain_mask_obj)
        mask_obj = rain_mask_obj 
        if cellsize is not None:
            if cellsize > rain_mask_obj.header['cellsize']:
                mask_obj = rain_mask_obj.grid_resample_nearest(cellsize)  
        ind_nan = np.isnan(mask_obj.array)
        rain_mask = mask_obj.array+0
        rain_mask[ind_nan] = 0
        mask_ind = rain_mask.flatten(order='F').astype('int64')
        rain_vect = cell_rain[mask_ind]
        rain_array = np.reshape(rain_vect, mask_obj.array.shape, order='F')
        rain_array[ind_nan] = np.nan
        rain_map_obj = Raster(array=rain_array, header=mask_obj.header)
    fig, ax = rain_map_obj.mapshow(cax_str=cax_str, **kwargs)
    xbound = ax.get_xbound()
    ybound = ax.get_ybound()
    # draw shape file on the rainfall map
    if shp_file is not None:
        sf = shapefile.Reader(shp_file)
        for shape in sf.shapeRecords():
            x = [i[0] for i in shape.shape.points[:]]
            y = [i[1] for i in shape.shape.points[:]]
            ax.plot(x, y, color='r', linewidth=1)
    ax.set_xbound(xbound)
    ax.set_ybound(ybound)
    if title is not None:
        ax.set_title(title)
    if figname is not None:
        fig.savefig(figname, dpi=dpi, bbox_inches = 'tight', pad_inches = 0.02)
        plt.close(fig)
    return rain_map_obj

def plot_time_series(plot_data=None, rain_source=None, rain_mask=None,
                     method='mean', start_date=None, datetime_interval=24, 
                     datetime_format='%m-%d', title_str=None,
                     **kwargs):
    """ Plot time series of average rainfall rate inside the model domain

    Args:
        method: 'mean'|'max','min','mean'method to calculate gridded rainfall over the model domain

    """
    if plot_data is None:
        plot_data = get_time_series(rain_source, rain_mask, start_date, method)
    time_x = plot_data[:,0]
    value_y = plot_data[:,1]
    fig, ax = plt.subplots()
    ax.plot(time_x, value_y, **kwargs)
    if 'start_date' in kwargs.keys():
        ax.xaxis.set_major_locator(mdates.HourLocator(
                interval=datetime_interval))
        ax.xaxis.set_major_formatter(mdates.DateFormatter(datetime_format))
    ax.set_ylabel(method+' rainfall rate (mm/h)')
    ax.grid(True)
    if title_str is None:
        title_str = method+' precipitation in the model domain'
    ax.set_title(title_str)
    plt.show()
    return fig, ax

def create_animation(output_file, rain_source, mask_file,
                     duration=0.5, **kwargs):
    """ Create animation of gridded rainfall rate

    Args:
        mask_header: (dict) header file provide georeference of rainfall mask
        start_date: a datetime object to give the initial date and time of rain
        duration: duration for each frame (seconds)
        cellsize: sclar (meter) the size of rainfall grid cells

    """
    fig_names = create_pictures(rain_source, mask_file, **kwargs)
    # create animation with the images
    images = []
    for fig_name in fig_names:
        images.append(imageio.imread(fig_name))
        os.remove(fig_name)
    # save animation and delete images
    if not output_file.endswith('.gif'):
        output_file = output_file+'.gif'
    imageio.mimsave(output_file, images, duration=duration)

def create_mp4(output_file, rain_source, mask_file, fps=10, **kwargs):
    fig_names = create_pictures(rain_source, mask_file, **kwargs)    
    if not output_file.endswith('.mp4'):
        output_file = output_file+'.mp4'
    print(output_file)
    writer = imageio.get_writer(output_file, 'MP4', fps=fps)
    for fig_name in fig_names:
        writer.append_data(imageio.imread(fig_name))
        os.remove(fig_name)
    writer.close()
    
def create_pictures(rain_source, mask_file, cellsize=1000, 
                    start_date=None, dpi=100, shp_file=None, **kwargs):
    """ create rainfall rate images

    Args:
        rain_source: a numpy array of rainfall time series, 1st column is time in seconds, other columns are rainfall intensity in m/s
        mask_file: a arc grid or a Raster object

    """
    time_series = rain_source[:,0]
    # create images
    fig_names = []
    for i in np.arange(0, time_series.size):
        print(str(i)+'/'+str(time_series.size))
        fig_name = 'temp'+str(i)+'.png'
        fig_names.append(fig_name)
        one_source = rain_source[i, :]
        one_source = one_source.reshape((1, one_source.size))
        if start_date is None:
            title_str = '{:.0f}'.format(time_series[i])+'s'
        else:
            title_str = start_date+timedelta(seconds=time_series[i])
            title_str = title_str.strftime("%m/%d/%Y %H:%M:%S")
        get_spatial_map(one_source, mask_file, cellsize=cellsize,
                        shp_file=shp_file, title=title_str,
                        figname=fig_name, dpi=dpi, **kwargs)
    return fig_names
    
def _check_rainfall_rate_values(rain_source, times_in_1st_col=True):
    """ Check the rainfall rate values in rain source array

    Args:
        times_in_1st_col: True|False indicate whether the first column is times

    Return:
        float: maximum rainfall rate in mm/h
        float: average rainfall rate in mm/h
        
    """
    # get the pure rainfall rate values
    if times_in_1st_col:
        rain_values = rain_source[:, 1:]
        time_series = rain_source[:, 0]
    else:
        rain_values = rain_source
        time_series = np.arange(rain_values.shape[0])
    # convert the unit of rain rate values from m/s to mm/h
    rain_values_mmh = rain_values*3600*1000
    values_max = rain_values_mmh.max()
    values_mean = rain_values.mean(axis=1)
    rain_total_amount = np.trapz(y=values_mean, x=time_series) # in meter
    duration = np.ptp(time_series)
    rain_rate_mean = rain_total_amount*1000/(duration/3600) #mm/h
    if values_max > 1000 or rain_rate_mean > 500:
        warnings.warn('Very large rainfall rates, better check your data!')
        print('Max rain: {:.2f} mm/h, Average rain: {:.2f} mm/h'.\
              format(values_max, rain_rate_mean))
    return values_max, rain_rate_mean



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Apr  1 14:56:15 2020
# Author: Xiaoding Ming

"""

Demo functions
==============

To do:
    * Create the demo

---------------

"""
import os
import pkg_resources
import numpy as np
from .InputHipims import InputHipims
from .OutputHipims import OutputHipims
#from hydro_raster.Raster import Raster

def demo_input(num_of_sections=1, set_example_inputs=True,
               figname=None, dpi=200, **kwargs):
    """ A demonstration to generate a hipims input object
    
    Args:
        set_example_inputs: (True|False) if True, initial condition, boundary 
            condition, rainfall source, and gauge postion will be set to the 
            input object according to sample data.
        figname: (string) if given, a domain map will saved

    """
    dem_file = pkg_resources.resource_filename(__name__,
                                             'sample/DEM.gz')
    obj_in = InputHipims(dem_data=dem_file, num_of_sections=num_of_sections)
    if set_example_inputs:
        __set_defaul_input(obj_in)
    # show model summary print(obj_in)
    obj_in.Summary.display()
    fig, ax = obj_in.domain_show(**kwargs)
    ax.set_title('Domain map')
    if figname is not None:
        fig.savefig(figname, dpi=dpi)
    return obj_in

def demo_output(case_folder, num_of_sections=1):
    """ A demonstration to generate a hipims output object

    Args:
        case_folder: (string) path to case folder
        num_of_sections: (int) number of domains
    
    Note:
        a input folder and files must be created before using this function
    """
    obj_out = OutputHipims(case_folder=case_folder,
                           num_of_sections=num_of_sections)
    return obj_out


def get_sample_data():
    """ Get sample data for demonstartion

    Returns:
        dem_file: string for DEM file name
        demo_data: a dictionary with boundary_condition, rain_source, and 
            gauges_pos data
        data_path: string the path of sample data
    """
    data_path = pkg_resources.resource_filename(__name__, 'sample')
    dem_file = os.path.join(data_path, 'DEM.gz')
    demo_data = {
        'boundary_condition': [
            {'polyPoints': np.array([[1427, 195],
                                 [1446, 243]]),
             'type': 'open',
             'hU': [[0, 30], [60, 20]]},
            {'polyPoints': np.array([[58, 1645],
                                 [72, 1170]]),
             'type': 'open',
             'h': [[0, 10], [60, 10]]}],
        'rain_source': np.array([[0, 2.77777778e-06],
                                 [1800, 2.77777778e-05],
                                 [3600, 0]]),
        'gauges_pos': np.array([[300, 1600],
                                [400, 1200]])
    }
    return dem_file, demo_data, data_path
    
# =============private functions==================
def __set_defaul_input(obj_in):
    """Set some default values for an InputHipims object
    """
    # load data for the demo
    _, demo_data, _ = get_sample_data()
    # define initial condition
    h0 = obj_in.DEM.array+0
    h0[np.isnan(h0)] = 0
    h0[h0 < 50] = 0
    h0[h0 >= 50] = 1
    # set initial water depth (h0) and velocity (hU0x, hU0y)
    obj_in.set_initial_condition('h0', h0)
    obj_in.set_initial_condition('hU0x', h0*0.0001)
    obj_in.set_initial_condition('hU0y', h0*0.0002)
    # define boundary condition
    bound_list = demo_data['boundary_condition']
    obj_in.set_boundary_condition(bound_list, outline_boundary='fall')
    # define and set rainfall mask and source (two rainfall sources)
    rain_source = demo_data['rain_source']
    obj_in.set_rainfall(rain_mask=0, rain_source=rain_source)
    # define and set monitor positions
    gauges_pos = demo_data['gauges_pos']
    obj_in.set_gauges_position(gauges_pos)



# -*- coding: utf-8 -*-
# @Author: Xiaodong Ming
# @Date:   2022-03-22 18:01:20
# @Last Modified by:   Xiaodong Ming
# @Last Modified time: 2022-03-23 14:56:34
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Wed Apr  1 23:46:54 2020
# Author: Xiaodong Ming
"""
OutputHipims
============

To do:
    Process output files from HiPIMS

----------------------

"""
import os
import datetime
import warnings
import numpy as np
import pandas as pd
from hydro_raster import spatial_analysis as sp
from hydro_raster.Raster import Raster
from .Summary import Summary
from .indep_functions import save_as_dict, _create_io_folders, load_object
class OutputHipims:
    """To read and analyze otuput files from a HiPIMS flood model

    Attributes:
        case_folder: (str) the absolute path of the case folder
        input_folder: (str|list of strings) the absolute path of the 
            input folder(s)
        output_folder: (str|list of strings) the absolute path of the 
            output folder(s)
        number_of_sections: (int) the number of subdomains of the model
        header: (dict or list of dict) provide the header information
        header_list: a list of sub headers [only for multi-gpu model]
        ref_datetime: 
        times_simu: (dataframe) with variable 'times' (simulated time in 
                    seconds) and 'date_times' if ref_datetime is defined.
        gauge_values_all: (dict) 'h', 'eta', 'hU', array with values of all 
            gauge positions, not time column
        gauge_values: gauge timeseries summarized from a series of gauge points
        grid_results: a list of Raster objects for gridded results

    """  
    def __init__(self, input_obj=None, case_folder=None,
                 num_of_sections=1, header_file_tag=None):
        """Initialize the object with a InputHiPIMS object or a case folder and
            the number of sections

        Args:
            header_file_tag: the output file to read grid header, e.g. 'h_0'
        """
        # pass argument values
        if input_obj is None:
            self.case_folder = case_folder
            self.num_of_sections = num_of_sections
            self._set_IO_folders()
            self._set_grid_header(asc_file=header_file_tag)
            summary_file = os.path.join(case_folder, 'readme.txt')
            summary_obj = Summary(summary_file)
            self.Summary = summary_obj
        elif type(input_obj) is dict: # load from a dict
            for key, value in input_obj.items():
                if key=='Summary':
                    self.Summary = Summary(value)
                else:
                    self.__dict__[key] = value
        elif hasattr(input_obj, 'DEM'):
            # get information from the input object
            # case_folder, num_of_sections, header
            self.case_folder = input_obj.get_case_folder()
            self.num_of_sections = input_obj.num_of_sections
            self.header = input_obj.header
            self.dem_array = input_obj.DEM.array
            self.crs = input_obj.DEM.crs
            self.output_folder = input_obj._data_folders['output']
            self.input_folder = input_obj._data_folders['input']
            self.Summary = input_obj.Summary
            if input_obj.num_of_sections>1:
                header_list = []
                output_folder = []
                input_folder = []
                for sub_obj in input_obj.Sections:
                    header_list.append(sub_obj.Raster.header)
                    output_folder.append(sub_obj._data_folders['output'])
                    input_folder.append(sub_obj._data_folders['input'])
                self.header_list = header_list  
                self.output_folder = output_folder
                self.input_folder = input_folder
        else:
            raise IOError('The first argument (input_obj) must be '
                          'a InputHipims object')

    def read_gauges_file(self, file_tag='h', compressed=False):
        """ Read gauges files for time seires of values at the monitored gauges

        Args:
            file_tag: h, hU, eta, corresponding to h_gauges.dat, hU_gauges.dat,
                and eta_gauges.dat, respectively

        Return:
            gauges_pos, the coordinates of gauges within the model domain

            times, time in seconds

            values, gauge values corresponding to the gauges position

        """
        if self.num_of_sections==1:
            output_folder = self.output_folder
            gauge_output_file = os.path.join(output_folder,
                                             file_tag+'_gauges.dat')
            gauge_pos_file = os.path.join(self.input_folder,
                                          'field', 'gauges_pos.dat')
            if compressed:
                gauge_output_file = gauge_output_file+'.gz'
                gauge_pos_file = gauge_pos_file+'.gz'
            times, values = _read_one_gauge_file(gauge_output_file)
            gauges_pos = np.loadtxt(gauge_pos_file, dtype='float64', ndmin=2)
        else: # multi-GPU
            gauges_value_all, gauges_pos = _combine_gauges_data_via_ind(
                    self.case_folder, self.num_of_sections, file_tag)
            if file_tag == 'hU':
                values_x = np.array(gauges_value_all[0].iloc[:,:-1])
                values_y = np.array(gauges_value_all[1].iloc[:,:-1])
                values = np.array([values_x, values_y])
                times = np.array(gauges_value_all[0]['times'])
            else:
                values = np.array(gauges_value_all.iloc[:,:-1])
                times = np.array(gauges_value_all['times'])
        self.times_simu = pd.DataFrame({'times':times})
        if hasattr(self, 'ref_datetime'):
            times_delta = times.astype('timedelta64[s]')
            date_times = np.datetime64(self.ref_datetime)+times_delta                    
            self.times_simu['date_times'] = date_times
        if not hasattr(self, 'gauge_values_all'):
            self.gauge_values_all = {}
        self.gauge_values_all[file_tag] = values
        self.gauges_pos = gauges_pos
        return gauges_pos, times, values
    
    def read_grid_file(self, file_tag='h_0', compressed=False):
        """Read asc grid files from output

        Return:
            Raster: a raster object of the grid

        """
        
        if not file_tag.endswith('.asc'):
            file_tag = file_tag+'.asc'
        if compressed:
            file_tag = file_tag+'.gz'
        if self.num_of_sections==1:
            file_name = os.path.join(self.output_folder, file_tag)
            print('reading file '+file_name)
            if os.path.isfile(file_name):
                grid_array, _, _ = sp.arcgridread(file_name)
            else:
                warnings.warn(file_name+' is not found!')
                grid_array = None
        else: # multi-GPU
            print('reading file '+file_tag)
            grid_array = self._combine_multi_gpu_grid_data(file_tag)
        if grid_array is None:
            warnings.warn(file_tag+' is not read successfully!')
            grid_obj = None
        else:
            grid_obj = Raster(array=grid_array, header=self.header)
            if hasattr(self, 'crs'):
                grid_obj.set_crs(self.crs)
        return grid_obj
    
    def read_velocity_files(self, time_tag):
        """
        read h, hUx and hUy files, calculate velocity and return a grid obj

        Parameters
        ----------
        time_tag : string
            time string in seconds, e.g. '7200'

        Returns
        -------
        velocity_ras : Raster object
            velocity values on the model grid.
        list: list of 2D array
            velocity values in x and y direction.

        """
        file_tag_x = 'hUx_'+time_tag
        file_tag_y = 'hUy_'+time_tag
        file_tag_h = 'h_'+time_tag
        grid_x = self.read_grid_file(file_tag=file_tag_x)
        grid_y = self.read_grid_file(file_tag=file_tag_y)
        grid_h = self.read_grid_file(file_tag=file_tag_h)
        array_h = grid_h.array+1e-6
        array_u = grid_x.array/array_h
        array_v = grid_y.array/array_h
        array_vel = np.sqrt(array_u**2+array_v**2)
        velocity_ras = Raster(array=array_vel, header=self.header)
        if hasattr(self, 'crs'):
            velocity_ras.set_crs(self.crs)
        return velocity_ras, [array_u, array_v]
    
    def add_gauge_results(self, var_name, gauge_name='All', gauge_ind=None,  
                          compressed=False):
        """ add simulated value to the object gauge by gauge

        Args:
            var_name: 'h', 'hU', 'eta'
            gauge_name: 'All' add all gauges, then gauge_ind not needed

        """
        # read all gauge data in all positions
        if not hasattr(self, 'gauge_values_all'):
            self.read_gauges_file(var_name, compressed)
        else:
            if var_name not in self.gauge_values_all.keys():
                self.read_gauges_file(var_name, compressed)
        if gauge_ind is not None:
            gauge_ind = np.array(gauge_ind, ndmin=2)
            # add position data for a gauge
            values = self.gauge_values_all[var_name]+0
            values_pd = self.times_simu.copy()
            if not hasattr(self, 'gauge_values'):
                self.gauge_values = {}
            if var_name == 'h': # calculation method is min
                one_gauge_v = values[:, gauge_ind]
                one_gauge_v = one_gauge_v.max(axis=1)
                values_pd['values'] = one_gauge_v
            elif var_name == 'hU':
                one_gauge_v = values[:, :, gauge_ind]
                one_gauge_v = one_gauge_v.sum(axis=2)*self.header['cellsize']
                values_pd['values_x'] = one_gauge_v[0]
                values_pd['values_y'] = one_gauge_v[1]
            elif var_name == 'eta':
                if gauge_ind.size > 1:
                    raise ValueError('gauge_ind for eta must be a scalar')
                else:
                    one_gauge_v = values[:, gauge_ind[0]]
                    values_pd['values'] = one_gauge_v
            if gauge_name in self.gauge_values.keys():
                gauge_dict = self.gauge_values[gauge_name]
                gauge_dict[var_name] = values_pd
            else:
                gauge_dict = {var_name:values_pd}
            self.gauge_values[gauge_name] = gauge_dict
    
    def add_grid_results(self, result_names='all', compressed=False):
        """Read and return Raster object to attribute 'grid_results'

        Args:
            result_names: string or list of string, gives the name of grid file
                default is to read and save all asc files

        """
        if not hasattr(self, 'grid_results'):
            self.grid_results = {}
        if type(result_names) is not list: # for a list of files
            if result_names=='all':
                times_array, t_max = self._read_run_time()
                result_names = ['h_max_'+'{:g}'.format(t_max)]
                for one_time in times_array:
                    result_names.append('h_'+'{:g}'.format(one_time))
                    result_names.append('hUx_'+'{:g}'.format(one_time))
                    result_names.append('hUy_'+'{:g}'.format(one_time))
            else:
                result_names = [result_names]
        if len(result_names)>1:
            print('To read files:')
            print(result_names)
        for file_tag in result_names:
            grid_obj = self.read_grid_file(file_tag, compressed)
            self.grid_results[file_tag] = grid_obj.array

    def set_ref_datetime(self, date_time, str_format='%Y-%m-%d %H:%M:%S'):
        """Set the refernce datetime of the simulation
        """
        if type(date_time) is str:
            self.ref_datetime = datetime.datetime.strptime(date_time,
                                                           str_format)
        elif type(date_time) is datetime.datetime:
            self.ref_datetime = date_time
        else:
            raise IOError('date_time must be a datetime object or a string')
    
    def _read_run_time(self):
        # read the times_setup file to get time strings of outputs
        if self.num_of_sections==1:
            time_file = os.path.join(self.case_folder, 'input', 
                                     'times_setup.dat')
        else:
            time_file = os.path.join(self.case_folder, 'times_setup.dat')
        time_array = np.loadtxt(time_file)
        t0 = time_array[0]
        t_max = time_array[1]
        dt = time_array[3]
        num_step = (t_max-t0)/dt+1
        num_step = int(num_step)
        times_array = np.linspace(0, 3600, num_step)
        return times_array, t_max
    
    def _combine_multi_gpu_grid_data(self, grid_file_name):
        """Combine multi-gpu grid files into a single file

        grid_file_name: string endswith '.asc' or '.txt'
        """
        header_global = self.header
        header_list = self.header_list
        output_folder = self.output_folder
        grid_shape = (header_global['nrows'], header_global['ncols'])
        array_global = np.zeros(grid_shape)
        # check file existence
        file_complete = True
        if grid_file_name == 'DEM.txt':
            input_folder = self.input_folder
            folder_list = [os.path.join(x, 'mesh') for x in input_folder]
        else:
            folder_list = output_folder
        for folder0 in folder_list:
            file_name = os.path.join(folder0, grid_file_name)
            if os.path.isfile(file_name):
                pass
            else:
                warnings.warn(file_name+' is not found!')
                file_complete = False
        if file_complete:
            for header0, folder0 in zip(header_list, folder_list):
                ind_top, ind_bottom = _header2row_numbers(header0, header_global)
                file_name = os.path.join(folder0, grid_file_name)
                array_local, _, _ = sp.arcgridread(file_name)
                array_global[ind_top:ind_bottom+1,:] = array_local
        else:
            array_global = None
        return array_global
    
    def _set_IO_folders(self):
        """ Set input and output folder/folders

        case_folder, num_of_sections are required
        """
        case_folder = self.case_folder
        num_of_sections = self.num_of_sections
        if num_of_sections == 1: # single gpu
            data_folders = _create_io_folders(case_folder)
            output_folder = data_folders['output']
            input_folder = data_folders['input']
        else: #multi-gpu model
            output_folder = []
            input_folder = []
            for i in range(num_of_sections):
                section_case_folder = os.path.join(case_folder, str(i))
                data_folders = _create_io_folders(section_case_folder)
                output_folder.append(data_folders['output'])  
                input_folder.append(data_folders['input'])
        self.output_folder = output_folder
        self.input_folder = input_folder
    
    def _set_grid_header(self, asc_file=None):
        """set header for the grid of the output object

        num_of_sections, input_folder, output_folder are required

        asc_file: the file to get grid header, default is DEM.txt

        """
        
        num_of_sections = self.num_of_sections
        output_folder = self.output_folder
        input_folder = self.input_folder
        warning_message = ('The grid header is not set, '
                           'try _set_grid_header(asc_file) to set a '
                           'header if it is required in your next steps')
        if num_of_sections == 1:
            if asc_file is None:
                file_name = os.path.join(input_folder, 'mesh', 'DEM.txt')
            else:
                file_name = os.path.join(output_folder, asc_file)
            if os.path.exists(file_name):
                dem_ras = Raster(file_name)
                self.header = dem_ras.header
                self.dem_array = dem_ras.array
            else:
                warnings.warn(warning_message)
        else: #multi-gpu model
            headers = []
            header_set = True
            for i in np.arange(num_of_sections):
                if asc_file is None:
                    file_name = os.path.join(input_folder[i],'mesh', 'DEM.txt')
                else:
                    file_name = os.path.join(output_folder[i], asc_file)
                if os.path.exists(file_name):
                    header = sp.arc_header_read(file_name)
                    headers.append(header)
                    header_set = True*header_set
                else:
                    warnings.warn(warning_message)
                    header_set = False
            if header_set:
                self.header_list = headers
                self.header = _header_local2global(headers)
            if asc_file is None:
                self.dem_array = self._combine_multi_gpu_grid_data('DEM.txt')
    
    def save_object(self, file_name):
        """Save the object to a pickle file
        """
#        save_object(self, file_name, compression=True)
        save_as_dict(self, file_name)
    
#%% =======================Supporting functions===============================
def load_output_object(filename):
    """load object from a dictionary and return as an OutputHipims object
    
    Args:
        filename: a string giving the object file name
    Return: 
        An object of OutputHipims
    """
    obj_dict = load_object(filename)
    if type(obj_dict) is OutputHipims:
        obj_out = obj_dict
    elif type(obj_dict) is dict:
        obj_out = OutputHipims(obj_dict)
    else:
        raise ValueError(filename+' should store either a dict or OutputHipims'
                         ' object')
    return obj_out
    
def _combine_gauges_data_via_ind(case_folder, num_section, file_tag):
    """Combine gauges outputs from multi-gpu models according to gauges

    position index.

    gauge_ind.dat

    gauge_pos.dat in each subdoamin input/field folder

    """
    # read index
    ind_list =[]
    ind_max = 0
    gauges_pos_all = []
    for i in range(num_section):
        gauge_ind_file = os.path.join(case_folder, str(i), 'input', 'field',
                                      'gauges_ind.dat')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ind_1 = np.loadtxt(gauge_ind_file, dtype='int')
        if ind_1.size>0:
            ind_max = max(ind_max, ind_1.max())
            file_name = os.path.join(case_folder, str(i), 'input', 'field',
                                     'gauges_pos.dat')
            gauge_xy = np.loadtxt(file_name, dtype='float64', ndmin=2)
            gauges_pos_all.append(gauge_xy)
        ind_list.append(ind_1)
    gauges_pos_all = np.concatenate(gauges_pos_all, axis=0)
    num_gauges = ind_max+1
    if file_tag == 'hU':
        gauges_value_x = pd.DataFrame(columns=np.arange(num_gauges))
        gauges_value_y = pd.DataFrame(columns=np.arange(num_gauges))
        gauges_value = [gauges_value_x, gauges_value_y]
    else:
        gauges_value = pd.DataFrame(columns=np.arange(num_gauges))
    for i in range(num_section):
        file_name = os.path.join(case_folder, str(i), 'output',
                                 file_tag+'_gauges.dat')
        if ind_list[i].size>0:
            t_value = np.loadtxt(file_name, dtype='float64', ndmin=2)
            values = t_value[:,1:]
            times = t_value[:, 0]
            if file_tag == 'hU':
                gauges_value_x['times'] = times
                gauges_value_y['times'] = times
                values_x = values[:, 0::2]
                values_y = values[:, 1::2]
                gauges_value_x.iloc[0:values.shape[0], ind_list[i]] = values_x
                gauges_value_y.iloc[0:values.shape[0], ind_list[i]] = values_y
            else:
                gauges_value['times'] = times
                gauges_value.iloc[0:values.shape[0], ind_list[i]] = values        
    return gauges_value, gauges_pos_all
    
def _combine_multi_gpu_gauges_data(header_list, case_folder, file_tag):
    """ Combine gauges outputs from multi-gpu models according to gauges

    position data.

    gauges_pos.dat for each domain must be available

    """
    gauges_array = []
    value_array = []
    for i in range(len(header_list)):
        domain_header = header_list[i]
        gauge_pos_file = os.path.join(case_folder, str(i), 'input', 'field',
                                      'gauges_pos.dat')
        gauge_xy = np.loadtxt(gauge_pos_file, dtype='float64', ndmin=2)
        if gauge_xy.size>=2:
            gauge_ind = _find_gauges_inside_domain(domain_header, gauge_xy)
            gauges_array.append(gauge_xy[gauge_ind,:])
            file_name = os.path.join(case_folder, str(i), 'output',
                                     file_tag+'_gauges.dat')
            times, values = _read_one_gauge_file(file_name, gauge_ind)
            value_array.append(values)
    gauges_array = np.concatenate(gauges_array, axis=0)
    gauges_array, ind = np.unique(gauges_array, axis=0, return_index=True)
    if values.ndim == 2:
        value_array = np.concatenate(value_array, axis=1)
        value_array = value_array[:, ind]
    else: # values.ndim == 3
        value_array = np.concatenate(value_array, axis=2)
        value_array = value_array[:, :, ind]
    return gauges_array, times, value_array

def _find_gauges_inside_domain(domain_header, gauge_xy):
    """ Find the gauges inside a domain

    domain_header: (dict) header of the domain DEM

    gauge_xy: xy coordinate of the gauges

    Return: (numpy array) index of gauges inside the model domain

    """
    extent = sp.header2extent(domain_header)
    ind_x = np.logical_and(gauge_xy[:, 0] > extent[0],
                           gauge_xy[:, 0] < extent[1])
    ind_y = np.logical_and(gauge_xy[:, 1] > extent[2],
                           gauge_xy[:, 1] < extent[3])
    gauge_ind = np.where(np.logical_and(ind_x, ind_y))
    gauge_ind = gauge_ind[0]
    return gauge_ind
        
def _read_one_gauge_file(file_name, gauge_ind=None):
    """ Read a gauge file and return time series and values with outside gauges
    removed

    Supporting function to read_gauges_file

    """
    t_value = np.loadtxt(file_name, dtype='float64', ndmin=2)
    times = t_value[:, 0]
    values = t_value[:, 1:]
    if 'hU_gauges.dat' in file_name:
        values = np.array([values[:, 0::2], values[:, 1::2]])
    if gauge_ind is not None:
        if values.ndim==2:
            values = values[:, gauge_ind]
        else: #ndim=3
            values = values[:, :, gauge_ind]
    return times, values

def _header2row_numbers(local_header, global_header):
    """Calculate local grid starting and ending rows in global grid

    Return:

        ind_top: the index of the top row

        ind_bottom: the index of the bottom row
        
    """
    y_bottom_gap = local_header['yllcorner']-global_header['yllcorner']
    row_gap = round(y_bottom_gap/local_header['cellsize'])
    ind_bottom = global_header['nrows']-1-row_gap
    ind_top = ind_bottom-local_header['nrows']+1
    ind_top = int(ind_top)
    ind_bottom = int(ind_bottom)
    return ind_top, ind_bottom
    
def _header_local2global(header_list):
    """Convert local headers to a global header
    """
    extent_array = []
    for header in header_list:
        extent = sp.header2extent(header)
        extent_array.append(extent)
    extent_array = np.asarray(extent_array)
    header_global = header_list[0].copy()
    y_bottom = extent_array[:,2].min()
    header_global['yllcorner'] = y_bottom
    y_top = extent_array[:,3].max()
    nrows = (y_top-y_bottom)/header_global['cellsize']
    header_global['nrows'] = int(round(nrows))
    return header_global



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
init
To do:
    initialize a package
Created on Wed Apr  1 14:56:15 2020

@author: Xiaodong Ming
"""
import numpy as np
from .demo_functions import demo_input
from .demo_functions import demo_output
from .demo_functions import get_sample_data
from .InputHipims import InputHipims, load_input_object
from .OutputHipims import OutputHipims, load_output_object
from .indep_functions import clean_output, _dict2grid
from .indep_functions import write_times_setup, write_device_setup, write_rain_source
from hydro_raster.Raster import Raster


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
setup the package
To do:
    [type the aims and objectives of this script]
Created on Wed Apr  1 15:03:58 2020

@author: Xiaodong Ming
"""
from setuptools import setup, find_packages
from os import path
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
      name='hipims_io',
      version='0.6.3',
      description='To process input and output files of the HiPIMS model',
      url='https://github.com/mingxiaodong/hipims_io_python',
      author='Xiaodong Ming',
      author_email='x.ming@lboro.ac.uk',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=['Development Status :: 3 - Alpha',
                   'Programming Language :: Python :: 3',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3',
                   'Operating System :: OS Independent',],
      keywords='HiPIMS model IO',
      license='LICENSE.txt',
      packages=find_packages(),
      include_package_data=True,
      package_data={'hipims_io': ['sample/Example_DEM.asc',
                                  'sample/DEM.gz',
                                  'sample/landcover.gz',
                                  'sample/rain_mask.gz',
                                  'sample/rain_source.csv',
                                  'sample/Example_data.npy'],},
      install_requires=['rasterio', 'scipy', 'pyshp', 'fiona', 'hydro_raster',
                        'matplotlib', 'numpy', 'pandas', 'imageio'],
      python_requires='>=3.6')

"""
Required python package
os, setuptools, copy, math, numpy, scipy, shutil, datetime, imageio, matplotlib,
mpl_toolkits, pkg_resources, gzip, pickle, warnings, pandas

#Sometimes you’ll want to use packages that are properly arranged with 
#setuptools, but aren’t published to PyPI. In those cases, you can specify a 
#list of one or more dependency_links URLs where the package can be downloaded,
#along with some additional hints, and setuptools will find and install the
#package correctly.
#For example, if a library is published on GitHub, you can specify it like:
setup(
    ...
    dependency_links=['http://github.com/user/repo/tarball/master#egg=package-1.0']
    ...
)
"""
